#!/usr/bin/env python3
# mypy: ignore-errors
"""Autogenerated TDF decoding logic"""

import ctypes

from infuse_iot.generated.tdf_base import TdfReadingBase, TdfStructBase


class structs:
    class tdf_struct_mcuboot_img_sem_ver(TdfStructBase):
        """MCUboot semantic versioning struct"""

        _fields_ = [
            ("major", ctypes.c_uint8),
            ("minor", ctypes.c_uint8),
            ("revision", ctypes.c_uint16),
            ("build_num", ctypes.c_uint32),
        ]
        _pack_ = 1
        _postfix_ = {
            "major": "",
            "minor": "",
            "revision": "",
            "build_num": "",
        }
        _display_fmt_ = {
            "major": "{}",
            "minor": "{}",
            "revision": "{}",
            "build_num": "0x{:08x}",
        }

    class tdf_struct_xyz_16bit(TdfStructBase):
        """Generic 3-axis sensor reading"""

        _fields_ = [
            ("x", ctypes.c_int16),
            ("y", ctypes.c_int16),
            ("z", ctypes.c_int16),
        ]
        _pack_ = 1
        _postfix_ = {
            "x": "",
            "y": "",
            "z": "",
        }
        _display_fmt_ = {
            "x": "{}",
            "y": "{}",
            "z": "{}",
        }

    class tdf_struct_gcs_location(TdfStructBase):
        """Geographic Coordinate System location"""

        _fields_ = [
            ("_latitude", ctypes.c_int32),
            ("_longitude", ctypes.c_int32),
            ("_height", ctypes.c_int32),
        ]
        _pack_ = 1
        _postfix_ = {
            "latitude": "deg",
            "longitude": "deg",
            "height": "m",
        }
        _display_fmt_ = {
            "latitude": "{:.5f}",
            "longitude": "{:.5f}",
            "height": "{:.3f}",
        }

        @property
        def latitude(self):
            return self._latitude * 1e-07

        @property
        def longitude(self):
            return self._longitude * 1e-07

        @property
        def height(self):
            return self._height * 0.001

    class tdf_struct_lte_cell_id_local(TdfStructBase):
        """LTE cell ID (Local)"""

        _fields_ = [
            ("eci", ctypes.c_uint32),
            ("tac", ctypes.c_uint16),
        ]
        _pack_ = 1
        _postfix_ = {
            "eci": "",
            "tac": "",
        }
        _display_fmt_ = {
            "eci": "{}",
            "tac": "{}",
        }

    class tdf_struct_lte_cell_id_global(TdfStructBase):
        """LTE cell ID (Global)"""

        _fields_ = [
            ("mcc", ctypes.c_uint16),
            ("mnc", ctypes.c_uint16),
            ("eci", ctypes.c_uint32),
            ("tac", ctypes.c_uint16),
        ]
        _pack_ = 1
        _postfix_ = {
            "mcc": "",
            "mnc": "",
            "eci": "",
            "tac": "",
        }
        _display_fmt_ = {
            "mcc": "{}",
            "mnc": "{}",
            "eci": "{}",
            "tac": "{}",
        }

    class tdf_struct_lte_cell_neighbour(TdfStructBase):
        """LTE cell ID (Global)"""

        _fields_ = [
            ("earfcn", ctypes.c_uint32),
            ("pci", ctypes.c_uint16),
            ("_time_diff", ctypes.c_uint16),
            ("_rsrp", ctypes.c_uint8),
            ("rsrq", ctypes.c_int8),
        ]
        _pack_ = 1
        _postfix_ = {
            "earfcn": "",
            "pci": "",
            "time_diff": "s",
            "rsrp": "dBm",
            "rsrq": "dB",
        }
        _display_fmt_ = {
            "earfcn": "{}",
            "pci": "{}",
            "time_diff": "{}",
            "rsrp": "{}",
            "rsrq": "{}",
        }

        @property
        def time_diff(self):
            return self._time_diff * 0.001

        @property
        def rsrp(self):
            return self._rsrp * -1

    class tdf_struct_bt_addr_le(TdfStructBase):
        """Bluetooth address type (bt_addr_le_t)"""

        _fields_ = [
            ("type", ctypes.c_uint8),
            ("_val", 6 * ctypes.c_uint8),
        ]
        _pack_ = 1
        _postfix_ = {
            "type": "",
            "val": "",
        }
        _display_fmt_ = {
            "type": "{}",
            "val": "0x{:012x}",
        }

        @property
        def val(self):
            return int.from_bytes(self._val, byteorder="little")

    class tdf_struct_eui48(TdfStructBase):
        """IEEE EUI-48 address"""

        _fields_ = [
            ("_val", 6 * ctypes.c_uint8),
        ]
        _pack_ = 1
        _postfix_ = {
            "val": "",
        }
        _display_fmt_ = {
            "val": "0x{:012x}",
        }

        @property
        def val(self):
            return int.from_bytes(self._val, byteorder="little")

    class tdf_struct_wifi_network_params(TdfStructBase):
        """WiFi network parameters"""

        _fields_ = [
            ("_bssid", 6 * ctypes.c_uint8),
            ("band", ctypes.c_uint8),
            ("channel", ctypes.c_uint8),
            ("iface_mode", ctypes.c_uint8),
            ("link_mode", ctypes.c_uint8),
            ("security", ctypes.c_uint8),
            ("rssi", ctypes.c_int8),
            ("beacon_interval", ctypes.c_uint16),
            ("twt_capable", ctypes.c_uint8),
        ]
        _pack_ = 1
        _postfix_ = {
            "bssid": "",
            "band": "",
            "channel": "",
            "iface_mode": "",
            "link_mode": "",
            "security": "",
            "rssi": "",
            "beacon_interval": "",
            "twt_capable": "",
        }
        _display_fmt_ = {
            "bssid": "0x{:012x}",
            "band": "{}",
            "channel": "{}",
            "iface_mode": "{}",
            "link_mode": "{}",
            "security": "{}",
            "rssi": "{}",
            "beacon_interval": "{}",
            "twt_capable": "{}",
        }

        @property
        def bssid(self):
            return int.from_bytes(self._bssid, byteorder="big")


class readings:
    class announce(TdfReadingBase):
        """Common announcement packet"""

        ID = 1
        NAME = "ANNOUNCE"
        _fields_ = [
            ("application", ctypes.c_uint32),
            ("version", structs.tdf_struct_mcuboot_img_sem_ver),
            ("kv_crc", ctypes.c_uint32),
            ("blocks", ctypes.c_uint32),
            ("uptime", ctypes.c_uint32),
            ("reboots", ctypes.c_uint16),
            ("flags", ctypes.c_uint8),
        ]
        _pack_ = 1
        _postfix_ = {
            "application": "",
            "version": "",
            "kv_crc": "",
            "blocks": "",
            "uptime": "",
            "reboots": "",
            "flags": "",
        }
        _display_fmt_ = {
            "application": "0x{:08x}",
            "version": "{}",
            "kv_crc": "0x{:08x}",
            "blocks": "{}",
            "uptime": "{}",
            "reboots": "{}",
            "flags": "0x{:02x}",
        }

    class battery_state(TdfReadingBase):
        """General battery state"""

        ID = 2
        NAME = "BATTERY_STATE"
        _fields_ = [
            ("voltage_mv", ctypes.c_uint32),
            ("current_ua", ctypes.c_int32),
            ("soc", ctypes.c_uint8),
        ]
        _pack_ = 1
        _postfix_ = {
            "voltage_mv": "mV",
            "current_ua": "uA",
            "soc": "%",
        }
        _display_fmt_ = {
            "voltage_mv": "{}",
            "current_ua": "{}",
            "soc": "{}",
        }

    class ambient_temp_pres_hum(TdfReadingBase):
        """Ambient temperature, pressure & humidity"""

        ID = 3
        NAME = "AMBIENT_TEMP_PRES_HUM"
        _fields_ = [
            ("_temperature", ctypes.c_int32),
            ("_pressure", ctypes.c_uint32),
            ("_humidity", ctypes.c_uint16),
        ]
        _pack_ = 1
        _postfix_ = {
            "temperature": "deg",
            "pressure": "kPA",
            "humidity": "%",
        }
        _display_fmt_ = {
            "temperature": "{:.3f}",
            "pressure": "{:.3f}",
            "humidity": "{:.2f}",
        }

        @property
        def temperature(self):
            return self._temperature * 0.001

        @property
        def pressure(self):
            return self._pressure * 0.001

        @property
        def humidity(self):
            return self._humidity * 0.01

    class ambient_temperature(TdfReadingBase):
        """Ambient temperature"""

        ID = 4
        NAME = "AMBIENT_TEMPERATURE"
        _fields_ = [
            ("_temperature", ctypes.c_int32),
        ]
        _pack_ = 1
        _postfix_ = {
            "temperature": "deg",
        }
        _display_fmt_ = {
            "temperature": "{:.3f}",
        }

        @property
        def temperature(self):
            return self._temperature * 0.001

    class time_sync(TdfReadingBase):
        """Time synchronised to new source"""

        ID = 5
        NAME = "TIME_SYNC"
        _fields_ = [
            ("source", ctypes.c_uint8),
            ("_shift", ctypes.c_int32),
        ]
        _pack_ = 1
        _postfix_ = {
            "source": "",
            "shift": "us",
        }
        _display_fmt_ = {
            "source": "{}",
            "shift": "{}",
        }

        @property
        def shift(self):
            return self._shift * 1e-06

    class reboot_info(TdfReadingBase):
        """Information pertaining to the previous reboot"""

        ID = 6
        NAME = "REBOOT_INFO"
        _fields_ = [
            ("reason", ctypes.c_uint8),
            ("hardware_flags", ctypes.c_uint32),
            ("count", ctypes.c_uint32),
            ("uptime", ctypes.c_uint32),
            ("param_1", ctypes.c_uint32),
            ("param_2", ctypes.c_uint32),
            ("thread", 8 * ctypes.c_char),
        ]
        _pack_ = 1
        _postfix_ = {
            "reason": "",
            "hardware_flags": "",
            "count": "",
            "uptime": "",
            "param_1": "",
            "param_2": "",
            "thread": "",
        }
        _display_fmt_ = {
            "reason": "{}",
            "hardware_flags": "0x{:08x}",
            "count": "{}",
            "uptime": "{}",
            "param_1": "0x{:08x}",
            "param_2": "0x{:08x}",
            "thread": "{}",
        }

    class announce_v2(TdfReadingBase):
        """Common announcement packet"""

        ID = 7
        NAME = "ANNOUNCE_V2"
        _fields_ = [
            ("application", ctypes.c_uint32),
            ("version", structs.tdf_struct_mcuboot_img_sem_ver),
            ("board_crc", ctypes.c_uint16),
            ("kv_crc", ctypes.c_uint32),
            ("blocks", ctypes.c_uint32),
            ("uptime", ctypes.c_uint32),
            ("reboots", ctypes.c_uint16),
            ("flags", ctypes.c_uint8),
        ]
        _pack_ = 1
        _postfix_ = {
            "application": "",
            "version": "",
            "board_crc": "",
            "kv_crc": "",
            "blocks": "",
            "uptime": "",
            "reboots": "",
            "flags": "",
        }
        _display_fmt_ = {
            "application": "0x{:08x}",
            "version": "{}",
            "board_crc": "0x{:04x}",
            "kv_crc": "0x{:08x}",
            "blocks": "{}",
            "uptime": "{}",
            "reboots": "{}",
            "flags": "0x{:02x}",
        }

    class acc_2g(TdfReadingBase):
        """Accelerometer +-2G"""

        ID = 10
        NAME = "ACC_2G"
        _fields_ = [
            ("sample", structs.tdf_struct_xyz_16bit),
        ]
        _pack_ = 1
        _postfix_ = {
            "sample": "",
        }
        _display_fmt_ = {
            "sample": "{}",
        }

    class acc_4g(TdfReadingBase):
        """Accelerometer +-4G"""

        ID = 11
        NAME = "ACC_4G"
        _fields_ = [
            ("sample", structs.tdf_struct_xyz_16bit),
        ]
        _pack_ = 1
        _postfix_ = {
            "sample": "",
        }
        _display_fmt_ = {
            "sample": "{}",
        }

    class acc_8g(TdfReadingBase):
        """Accelerometer +-8G"""

        ID = 12
        NAME = "ACC_8G"
        _fields_ = [
            ("sample", structs.tdf_struct_xyz_16bit),
        ]
        _pack_ = 1
        _postfix_ = {
            "sample": "",
        }
        _display_fmt_ = {
            "sample": "{}",
        }

    class acc_16g(TdfReadingBase):
        """Accelerometer +-16G"""

        ID = 13
        NAME = "ACC_16G"
        _fields_ = [
            ("sample", structs.tdf_struct_xyz_16bit),
        ]
        _pack_ = 1
        _postfix_ = {
            "sample": "",
        }
        _display_fmt_ = {
            "sample": "{}",
        }

    class gyr_125dps(TdfReadingBase):
        """Gyroscope +-125 DPS"""

        ID = 14
        NAME = "GYR_125DPS"
        _fields_ = [
            ("sample", structs.tdf_struct_xyz_16bit),
        ]
        _pack_ = 1
        _postfix_ = {
            "sample": "",
        }
        _display_fmt_ = {
            "sample": "{}",
        }

    class gyr_250dps(TdfReadingBase):
        """Gyroscope +-250 DPS"""

        ID = 15
        NAME = "GYR_250DPS"
        _fields_ = [
            ("sample", structs.tdf_struct_xyz_16bit),
        ]
        _pack_ = 1
        _postfix_ = {
            "sample": "",
        }
        _display_fmt_ = {
            "sample": "{}",
        }

    class gyr_500dps(TdfReadingBase):
        """Gyroscope +-500 DPS"""

        ID = 16
        NAME = "GYR_500DPS"
        _fields_ = [
            ("sample", structs.tdf_struct_xyz_16bit),
        ]
        _pack_ = 1
        _postfix_ = {
            "sample": "",
        }
        _display_fmt_ = {
            "sample": "{}",
        }

    class gyr_1000dps(TdfReadingBase):
        """Gyroscope +-1000 DPS"""

        ID = 17
        NAME = "GYR_1000DPS"
        _fields_ = [
            ("sample", structs.tdf_struct_xyz_16bit),
        ]
        _pack_ = 1
        _postfix_ = {
            "sample": "",
        }
        _display_fmt_ = {
            "sample": "{}",
        }

    class gyr_2000dps(TdfReadingBase):
        """Gyroscope +-2000 DPS"""

        ID = 18
        NAME = "GYR_2000DPS"
        _fields_ = [
            ("sample", structs.tdf_struct_xyz_16bit),
        ]
        _pack_ = 1
        _postfix_ = {
            "sample": "",
        }
        _display_fmt_ = {
            "sample": "{}",
        }

    class gcs_wgs84_llha(TdfReadingBase):
        """Geo-location (WGS-84) + accuracy"""

        ID = 19
        NAME = "GCS_WGS84_LLHA"
        _fields_ = [
            ("location", structs.tdf_struct_gcs_location),
            ("_h_acc", ctypes.c_int32),
            ("_v_acc", ctypes.c_int32),
        ]
        _pack_ = 1
        _postfix_ = {
            "location": "",
            "h_acc": "m",
            "v_acc": "m",
        }
        _display_fmt_ = {
            "location": "{}",
            "h_acc": "{:.3f}",
            "v_acc": "{:.3f}",
        }

        @property
        def h_acc(self):
            return self._h_acc * 0.001

        @property
        def v_acc(self):
            return self._v_acc * 0.001

    class ubx_nav_pvt(TdfReadingBase):
        """u-blox GNSS NAV-PVT message"""

        ID = 20
        NAME = "UBX_NAV_PVT"
        _fields_ = [
            ("itow", ctypes.c_uint32),
            ("year", ctypes.c_uint16),
            ("month", ctypes.c_uint8),
            ("day", ctypes.c_uint8),
            ("hour", ctypes.c_uint8),
            ("min", ctypes.c_uint8),
            ("sec", ctypes.c_uint8),
            ("valid", ctypes.c_uint8),
            ("t_acc", ctypes.c_uint32),
            ("nano", ctypes.c_int32),
            ("fix_type", ctypes.c_uint8),
            ("flags", ctypes.c_uint8),
            ("flags2", ctypes.c_uint8),
            ("num_sv", ctypes.c_uint8),
            ("_lon", ctypes.c_int32),
            ("_lat", ctypes.c_int32),
            ("_height", ctypes.c_int32),
            ("_h_msl", ctypes.c_int32),
            ("_h_acc", ctypes.c_uint32),
            ("_v_acc", ctypes.c_uint32),
            ("_vel_n", ctypes.c_int32),
            ("_vel_e", ctypes.c_int32),
            ("_vel_d", ctypes.c_int32),
            ("_g_speed", ctypes.c_int32),
            ("_head_mot", ctypes.c_int32),
            ("_s_acc", ctypes.c_uint32),
            ("_head_acc", ctypes.c_uint32),
            ("_p_dop", ctypes.c_uint16),
            ("flags3", ctypes.c_uint16),
            ("reserved0", 4 * ctypes.c_uint8),
            ("_head_veh", ctypes.c_int32),
            ("_mag_dec", ctypes.c_int16),
            ("_mag_acc", ctypes.c_uint16),
        ]
        _pack_ = 1
        _postfix_ = {
            "itow": "",
            "year": "",
            "month": "",
            "day": "",
            "hour": "",
            "min": "",
            "sec": "",
            "valid": "",
            "t_acc": "ns",
            "nano": "ns",
            "fix_type": "",
            "flags": "",
            "flags2": "",
            "num_sv": "",
            "lon": "deg",
            "lat": "deg",
            "height": "m",
            "h_msl": "m",
            "h_acc": "m",
            "v_acc": "m",
            "vel_n": "m/s",
            "vel_e": "m/s",
            "vel_d": "m/s",
            "g_speed": "m/s",
            "head_mot": "deg",
            "s_acc": "m/s",
            "head_acc": "deg",
            "p_dop": "",
            "flags3": "",
            "reserved0": "",
            "head_veh": "deg",
            "mag_dec": "deg",
            "mag_acc": "deg",
        }
        _display_fmt_ = {
            "itow": "{}",
            "year": "{}",
            "month": "{}",
            "day": "{}",
            "hour": "{}",
            "min": "{}",
            "sec": "{}",
            "valid": "0x{:02x}",
            "t_acc": "{}",
            "nano": "{}",
            "fix_type": "{}",
            "flags": "0x{:02x}",
            "flags2": "0x{:02x}",
            "num_sv": "{}",
            "lon": "{}",
            "lat": "{}",
            "height": "{}",
            "h_msl": "{}",
            "h_acc": "{}",
            "v_acc": "{}",
            "vel_n": "{}",
            "vel_e": "{}",
            "vel_d": "{}",
            "g_speed": "{}",
            "head_mot": "{}",
            "s_acc": "{}",
            "head_acc": "{}",
            "p_dop": "{}",
            "flags3": "0x{:04x}",
            "reserved0": "{}",
            "head_veh": "{}",
            "mag_dec": "{}",
            "mag_acc": "{}",
        }

        @property
        def lon(self):
            return self._lon * 1e-07

        @property
        def lat(self):
            return self._lat * 1e-07

        @property
        def height(self):
            return self._height * 0.001

        @property
        def h_msl(self):
            return self._h_msl * 0.001

        @property
        def h_acc(self):
            return self._h_acc * 0.001

        @property
        def v_acc(self):
            return self._v_acc * 0.001

        @property
        def vel_n(self):
            return self._vel_n * 0.001

        @property
        def vel_e(self):
            return self._vel_e * 0.001

        @property
        def vel_d(self):
            return self._vel_d * 0.001

        @property
        def g_speed(self):
            return self._g_speed * 0.001

        @property
        def head_mot(self):
            return self._head_mot * 1e-05

        @property
        def s_acc(self):
            return self._s_acc * 0.001

        @property
        def head_acc(self):
            return self._head_acc * 1e-05

        @property
        def p_dop(self):
            return self._p_dop * 0.01

        @property
        def head_veh(self):
            return self._head_veh * 1e-05

        @property
        def mag_dec(self):
            return self._mag_dec * 0.01

        @property
        def mag_acc(self):
            return self._mag_acc * 0.01

    class lte_conn_status(TdfReadingBase):
        """Information on service cell and registration status"""

        ID = 21
        NAME = "LTE_CONN_STATUS"
        _fields_ = [
            ("cell", structs.tdf_struct_lte_cell_id_global),
            ("earfcn", ctypes.c_uint32),
            ("status", ctypes.c_uint8),
            ("tech", ctypes.c_uint8),
            ("_rsrp", ctypes.c_uint8),
            ("rsrq", ctypes.c_int8),
        ]
        _pack_ = 1
        _postfix_ = {
            "cell": "",
            "earfcn": "",
            "status": "",
            "tech": "",
            "rsrp": "dBm",
            "rsrq": "dB",
        }
        _display_fmt_ = {
            "cell": "{}",
            "earfcn": "{}",
            "status": "{}",
            "tech": "{}",
            "rsrp": "{}",
            "rsrq": "{}",
        }

        @property
        def rsrp(self):
            return self._rsrp * -1

    class globalstar_pkt(TdfReadingBase):
        """9 byte payload transmitted over the Globalstar Simplex network"""

        ID = 22
        NAME = "GLOBALSTAR_PKT"
        _fields_ = [
            ("payload", 9 * ctypes.c_uint8),
        ]
        _pack_ = 1
        _postfix_ = {
            "payload": "",
        }
        _display_fmt_ = {
            "payload": "{}",
        }

    class acc_magnitude_std_dev(TdfReadingBase):
        """Accelerometer magnitude standard deviation over a window"""

        ID = 23
        NAME = "ACC_MAGNITUDE_STD_DEV"
        _fields_ = [
            ("count", ctypes.c_uint32),
            ("std_dev", ctypes.c_uint32),
        ]
        _pack_ = 1
        _postfix_ = {
            "count": "",
            "std_dev": "",
        }
        _display_fmt_ = {
            "count": "{}",
            "std_dev": "{}",
        }

    class activity_metric(TdfReadingBase):
        """Generic activity metric"""

        ID = 24
        NAME = "ACTIVITY_METRIC"
        _fields_ = [
            ("value", ctypes.c_uint32),
        ]
        _pack_ = 1
        _postfix_ = {
            "value": "",
        }
        _display_fmt_ = {
            "value": "{}",
        }

    class algorithm_output(TdfReadingBase):
        """Instantaneous algorithm output"""

        ID = 25
        NAME = "ALGORITHM_OUTPUT"
        _fields_ = [
            ("algorithm_id", ctypes.c_uint32),
            ("algorithm_version", ctypes.c_uint16),
            ("output", 0 * ctypes.c_uint8),
        ]
        _pack_ = 1
        _postfix_ = {
            "algorithm_id": "",
            "algorithm_version": "",
            "output": "",
        }
        _display_fmt_ = {
            "algorithm_id": "0x{:08x}",
            "algorithm_version": "{}",
            "output": "{}",
        }

    class runtime_error(TdfReadingBase):
        """Runtime error logging"""

        ID = 26
        NAME = "RUNTIME_ERROR"
        _fields_ = [
            ("error_id", ctypes.c_uint32),
            ("error_ctx", ctypes.c_uint32),
        ]
        _pack_ = 1
        _postfix_ = {
            "error_id": "",
            "error_ctx": "",
        }
        _display_fmt_ = {
            "error_id": "{}",
            "error_ctx": "{}",
        }

    class charger_en_control(TdfReadingBase):
        """Battery charging enable state"""

        ID = 27
        NAME = "CHARGER_EN_CONTROL"
        _fields_ = [
            ("enabled", ctypes.c_uint8),
        ]
        _pack_ = 1
        _postfix_ = {
            "enabled": "",
        }
        _display_fmt_ = {
            "enabled": "{}",
        }

    class gnss_fix_info(TdfReadingBase):
        """Metadata about a GNSS location fix"""

        ID = 28
        NAME = "GNSS_FIX_INFO"
        _fields_ = [
            ("time_fix", ctypes.c_uint16),
            ("location_fix", ctypes.c_uint16),
            ("num_sv", ctypes.c_uint8),
        ]
        _pack_ = 1
        _postfix_ = {
            "time_fix": "",
            "location_fix": "",
            "num_sv": "",
        }
        _display_fmt_ = {
            "time_fix": "{}",
            "location_fix": "{}",
            "num_sv": "{}",
        }

    class bluetooth_connection(TdfReadingBase):
        """Bluetooth connection state change"""

        ID = 29
        NAME = "BLUETOOTH_CONNECTION"
        _fields_ = [
            ("address", structs.tdf_struct_bt_addr_le),
            ("connected", ctypes.c_uint8),
        ]
        _pack_ = 1
        _postfix_ = {
            "address": "",
            "connected": "",
        }
        _display_fmt_ = {
            "address": "{}",
            "connected": "{}",
        }

    class bluetooth_rssi(TdfReadingBase):
        """Received signal strength of Bluetooth device"""

        ID = 30
        NAME = "BLUETOOTH_RSSI"
        _fields_ = [
            ("address", structs.tdf_struct_bt_addr_le),
            ("rssi", ctypes.c_int8),
        ]
        _pack_ = 1
        _postfix_ = {
            "address": "",
            "rssi": "dBm",
        }
        _display_fmt_ = {
            "address": "{}",
            "rssi": "{}",
        }

    class bluetooth_data_throughput(TdfReadingBase):
        """Data throughput of Bluetooth link"""

        ID = 31
        NAME = "BLUETOOTH_DATA_THROUGHPUT"
        _fields_ = [
            ("address", structs.tdf_struct_bt_addr_le),
            ("throughput", ctypes.c_int32),
        ]
        _pack_ = 1
        _postfix_ = {
            "address": "",
            "throughput": "B/sec",
        }
        _display_fmt_ = {
            "address": "{}",
            "throughput": "{}",
        }

    class algorithm_class_histogram(TdfReadingBase):
        """Algorithm output class histogram over a time window"""

        ID = 32
        NAME = "ALGORITHM_CLASS_HISTOGRAM"
        _fields_ = [
            ("algorithm_id", ctypes.c_uint32),
            ("algorithm_version", ctypes.c_uint16),
            ("classes", 0 * ctypes.c_uint8),
        ]
        _pack_ = 1
        _postfix_ = {
            "algorithm_id": "",
            "algorithm_version": "",
            "classes": "",
        }
        _display_fmt_ = {
            "algorithm_id": "0x{:08x}",
            "algorithm_version": "{}",
            "classes": "{}",
        }

    class algorithm_class_time_series(TdfReadingBase):
        """Algorithm output class time series vector"""

        ID = 33
        NAME = "ALGORITHM_CLASS_TIME_SERIES"
        _fields_ = [
            ("algorithm_id", ctypes.c_uint32),
            ("algorithm_version", ctypes.c_uint16),
            ("values", 0 * ctypes.c_uint8),
        ]
        _pack_ = 1
        _postfix_ = {
            "algorithm_id": "",
            "algorithm_version": "",
            "values": "",
        }
        _display_fmt_ = {
            "algorithm_id": "0x{:08x}",
            "algorithm_version": "{}",
            "values": "{}",
        }

    class lte_tac_cells(TdfReadingBase):
        """Information on cells in a tracking area"""

        ID = 34
        NAME = "LTE_TAC_CELLS"
        _fields_ = [
            ("cell", structs.tdf_struct_lte_cell_id_global),
            ("earfcn", ctypes.c_uint32),
            ("_rsrp", ctypes.c_uint8),
            ("rsrq", ctypes.c_int8),
            ("neighbours", 0 * structs.tdf_struct_lte_cell_neighbour),
        ]
        _pack_ = 1
        _postfix_ = {
            "cell": "",
            "earfcn": "",
            "rsrp": "dBm",
            "rsrq": "dB",
            "neighbours": "",
        }
        _display_fmt_ = {
            "cell": "{}",
            "earfcn": "{}",
            "rsrp": "{}",
            "rsrq": "{}",
            "neighbours": "{}",
        }

        @property
        def rsrp(self):
            return self._rsrp * -1

    class wifi_ap_info(TdfReadingBase):
        """Wi-Fi access point information"""

        ID = 35
        NAME = "WIFI_AP_INFO"
        _fields_ = [
            ("bssid", structs.tdf_struct_eui48),
            ("channel", ctypes.c_uint8),
            ("rsrp", ctypes.c_int8),
        ]
        _pack_ = 1
        _postfix_ = {
            "bssid": "",
            "channel": "",
            "rsrp": "",
        }
        _display_fmt_ = {
            "bssid": "{}",
            "channel": "{}",
            "rsrp": "{}",
        }

    class device_tilt(TdfReadingBase):
        """Tilt angle of the device"""

        ID = 36
        NAME = "DEVICE_TILT"
        _fields_ = [
            ("cosine", ctypes.c_float),
        ]
        _pack_ = 1
        _postfix_ = {
            "cosine": "",
        }
        _display_fmt_ = {
            "cosine": "{:.6f}",
        }

    class nrf9x_gnss_pvt(TdfReadingBase):
        """nRF9x GNSS PVT data frame"""

        ID = 37
        NAME = "NRF9X_GNSS_PVT"
        _fields_ = [
            ("_lat", ctypes.c_int32),
            ("_lon", ctypes.c_int32),
            ("_height", ctypes.c_int32),
            ("_h_acc", ctypes.c_uint32),
            ("_v_acc", ctypes.c_uint32),
            ("_h_speed", ctypes.c_int32),
            ("_h_speed_acc", ctypes.c_uint32),
            ("_v_speed", ctypes.c_int32),
            ("_v_speed_acc", ctypes.c_uint32),
            ("_head_mot", ctypes.c_int32),
            ("_head_acc", ctypes.c_uint32),
            ("year", ctypes.c_uint16),
            ("month", ctypes.c_uint8),
            ("day", ctypes.c_uint8),
            ("hour", ctypes.c_uint8),
            ("min", ctypes.c_uint8),
            ("sec", ctypes.c_uint8),
            ("ms", ctypes.c_uint16),
            ("_p_dop", ctypes.c_uint16),
            ("_h_dop", ctypes.c_uint16),
            ("_v_dop", ctypes.c_uint16),
            ("_t_dop", ctypes.c_uint16),
            ("flags", ctypes.c_uint8),
            ("num_sv", ctypes.c_uint8),
        ]
        _pack_ = 1
        _postfix_ = {
            "lat": "deg",
            "lon": "deg",
            "height": "m",
            "h_acc": "m",
            "v_acc": "m",
            "h_speed": "m/s",
            "h_speed_acc": "m/s",
            "v_speed": "m/s",
            "v_speed_acc": "m/s",
            "head_mot": "deg",
            "head_acc": "deg",
            "year": "",
            "month": "",
            "day": "",
            "hour": "",
            "min": "",
            "sec": "",
            "ms": "",
            "p_dop": "",
            "h_dop": "",
            "v_dop": "",
            "t_dop": "",
            "flags": "",
            "num_sv": "",
        }
        _display_fmt_ = {
            "lat": "{}",
            "lon": "{}",
            "height": "{}",
            "h_acc": "{}",
            "v_acc": "{}",
            "h_speed": "{}",
            "h_speed_acc": "{}",
            "v_speed": "{}",
            "v_speed_acc": "{}",
            "head_mot": "{}",
            "head_acc": "{}",
            "year": "{}",
            "month": "{}",
            "day": "{}",
            "hour": "{}",
            "min": "{}",
            "sec": "{}",
            "ms": "{}",
            "p_dop": "{}",
            "h_dop": "{}",
            "v_dop": "{}",
            "t_dop": "{}",
            "flags": "0x{:02x}",
            "num_sv": "{}",
        }

        @property
        def lat(self):
            return self._lat * 1e-07

        @property
        def lon(self):
            return self._lon * 1e-07

        @property
        def height(self):
            return self._height * 0.001

        @property
        def h_acc(self):
            return self._h_acc * 0.001

        @property
        def v_acc(self):
            return self._v_acc * 0.001

        @property
        def h_speed(self):
            return self._h_speed * 0.001

        @property
        def h_speed_acc(self):
            return self._h_speed_acc * 0.001

        @property
        def v_speed(self):
            return self._v_speed * 0.001

        @property
        def v_speed_acc(self):
            return self._v_speed_acc * 0.001

        @property
        def head_mot(self):
            return self._head_mot * 1e-05

        @property
        def head_acc(self):
            return self._head_acc * 1e-05

        @property
        def p_dop(self):
            return self._p_dop * 0.01

        @property
        def h_dop(self):
            return self._h_dop * 0.01

        @property
        def v_dop(self):
            return self._v_dop * 0.01

        @property
        def t_dop(self):
            return self._t_dop * 0.01

    class battery_charge_accumulated(TdfReadingBase):
        """Battery charge accumulated over time (+ve entering battery, -ve exiting battery)"""

        ID = 38
        NAME = "BATTERY_CHARGE_ACCUMULATED"
        _fields_ = [
            ("charge", ctypes.c_int32),
        ]
        _pack_ = 1
        _postfix_ = {
            "charge": "uAs",
        }
        _display_fmt_ = {
            "charge": "{}",
        }

    class infuse_bluetooth_rssi(TdfReadingBase):
        """Received signal strength of Infuse-IoT Bluetooth device"""

        ID = 39
        NAME = "INFUSE_BLUETOOTH_RSSI"
        _fields_ = [
            ("infuse_id", ctypes.c_uint64),
            ("rssi", ctypes.c_int8),
        ]
        _pack_ = 1
        _postfix_ = {
            "infuse_id": "",
            "rssi": "dBm",
        }
        _display_fmt_ = {
            "infuse_id": "{}",
            "rssi": "{}",
        }

    class adc_raw_8(TdfReadingBase):
        """Generic 8bit raw ADC reading"""

        ID = 40
        NAME = "ADC_RAW_8"
        _fields_ = [
            ("val", ctypes.c_int8),
        ]
        _pack_ = 1
        _postfix_ = {
            "val": "",
        }
        _display_fmt_ = {
            "val": "{}",
        }

    class adc_raw_16(TdfReadingBase):
        """Generic 16bit raw ADC reading"""

        ID = 41
        NAME = "ADC_RAW_16"
        _fields_ = [
            ("val", ctypes.c_int16),
        ]
        _pack_ = 1
        _postfix_ = {
            "val": "",
        }
        _display_fmt_ = {
            "val": "{}",
        }

    class adc_raw_32(TdfReadingBase):
        """Generic 32bit raw ADC reading"""

        ID = 42
        NAME = "ADC_RAW_32"
        _fields_ = [
            ("val", ctypes.c_int32),
        ]
        _pack_ = 1
        _postfix_ = {
            "val": "",
        }
        _display_fmt_ = {
            "val": "{}",
        }

    class annotation(TdfReadingBase):
        """Generic event annotation"""

        ID = 43
        NAME = "ANNOTATION"
        _fields_ = [
            ("timestamp", ctypes.c_uint32),
            ("event", 0 * ctypes.c_char),
        ]
        _pack_ = 1
        _postfix_ = {
            "timestamp": "",
            "event": "",
        }
        _display_fmt_ = {
            "timestamp": "{}",
            "event": "{}",
        }

    class lora_rx(TdfReadingBase):
        """Received LoRa packet"""

        ID = 44
        NAME = "LORA_RX"
        _fields_ = [
            ("snr", ctypes.c_int8),
            ("rssi", ctypes.c_int16),
            ("payload", 0 * ctypes.c_uint8),
        ]
        _pack_ = 1
        _postfix_ = {
            "snr": "dB",
            "rssi": "dBm",
            "payload": "",
        }
        _display_fmt_ = {
            "snr": "{}",
            "rssi": "{}",
            "payload": "0x{:02x}",
        }

    class lora_tx(TdfReadingBase):
        """Transmitted LoRa packet"""

        ID = 45
        NAME = "LORA_TX"
        _fields_ = [
            ("payload", 0 * ctypes.c_uint8),
        ]
        _pack_ = 1
        _postfix_ = {
            "payload": "",
        }
        _display_fmt_ = {
            "payload": "0x{:02x}",
        }

    class idx_array_freq(TdfReadingBase):
        """Sample frequency metadata for a TDF_DATA_FORMAT_IDX_ARRAY array"""

        ID = 46
        NAME = "IDX_ARRAY_FREQ"
        _fields_ = [
            ("tdf_id", ctypes.c_uint16),
            ("frequency", ctypes.c_uint32),
        ]
        _pack_ = 1
        _postfix_ = {
            "tdf_id": "",
            "frequency": "Hz",
        }
        _display_fmt_ = {
            "tdf_id": "{}",
            "frequency": "{}",
        }

    class idx_array_period(TdfReadingBase):
        """Sample frequency metadata for a TDF_DATA_FORMAT_IDX_ARRAY array"""

        ID = 47
        NAME = "IDX_ARRAY_PERIOD"
        _fields_ = [
            ("tdf_id", ctypes.c_uint16),
            ("period", ctypes.c_uint32),
        ]
        _pack_ = 1
        _postfix_ = {
            "tdf_id": "",
            "period": "ns",
        }
        _display_fmt_ = {
            "tdf_id": "{}",
            "period": "{}",
        }

    class wifi_connected(TdfReadingBase):
        """WiFi network is now connected"""

        ID = 48
        NAME = "WIFI_CONNECTED"
        _fields_ = [
            ("network", structs.tdf_struct_wifi_network_params),
        ]
        _pack_ = 1
        _postfix_ = {
            "network": "",
        }
        _display_fmt_ = {
            "network": "{}",
        }

    class wifi_connection_failed(TdfReadingBase):
        """Failed to connect to a WiFi network"""

        ID = 49
        NAME = "WIFI_CONNECTION_FAILED"
        _fields_ = [
            ("reason", ctypes.c_uint8),
        ]
        _pack_ = 1
        _postfix_ = {
            "reason": "",
        }
        _display_fmt_ = {
            "reason": "{}",
        }

    class wifi_disconnected(TdfReadingBase):
        """Wi-Fi network is now disconnected"""

        ID = 50
        NAME = "WIFI_DISCONNECTED"
        _fields_ = [
            ("reason", ctypes.c_uint8),
        ]
        _pack_ = 1
        _postfix_ = {
            "reason": "",
        }
        _display_fmt_ = {
            "reason": "{}",
        }

    class network_scan_count(TdfReadingBase):
        """Counts associated with scanning Wi-Fi and LTE networks"""

        ID = 51
        NAME = "NETWORK_SCAN_COUNT"
        _fields_ = [
            ("num_wifi", ctypes.c_uint8),
            ("num_lte", ctypes.c_uint8),
        ]
        _pack_ = 1
        _postfix_ = {
            "num_wifi": "",
            "num_lte": "",
        }
        _display_fmt_ = {
            "num_wifi": "{}",
            "num_lte": "{}",
        }

    class exception_stack_frame(TdfReadingBase):
        """Generic exception stack frame"""

        ID = 52
        NAME = "EXCEPTION_STACK_FRAME"
        _fields_ = [
            ("frame", 0 * ctypes.c_uint32),
        ]
        _pack_ = 1
        _postfix_ = {
            "frame": "",
        }
        _display_fmt_ = {
            "frame": "0x{:08x}",
        }

    class battery_voltage(TdfReadingBase):
        """Battery voltage"""

        ID = 53
        NAME = "BATTERY_VOLTAGE"
        _fields_ = [
            ("voltage", ctypes.c_uint16),
        ]
        _pack_ = 1
        _postfix_ = {
            "voltage": "mV",
        }
        _display_fmt_ = {
            "voltage": "{}",
        }

    class battery_soc(TdfReadingBase):
        """Battery state of charge"""

        ID = 54
        NAME = "BATTERY_SOC"
        _fields_ = [
            ("soc", ctypes.c_uint8),
        ]
        _pack_ = 1
        _postfix_ = {
            "soc": "%",
        }
        _display_fmt_ = {
            "soc": "{}",
        }

    class state_event_set(TdfReadingBase):
        """Infuse-IoT application state transitioned from cleared to set"""

        ID = 55
        NAME = "STATE_EVENT_SET"
        _fields_ = [
            ("state", ctypes.c_uint8),
        ]
        _pack_ = 1
        _postfix_ = {
            "state": "",
        }
        _display_fmt_ = {
            "state": "{}",
        }

    class state_event_cleared(TdfReadingBase):
        """Infuse-IoT application state transitioned from set to cleared"""

        ID = 56
        NAME = "STATE_EVENT_CLEARED"
        _fields_ = [
            ("state", ctypes.c_uint8),
        ]
        _pack_ = 1
        _postfix_ = {
            "state": "",
        }
        _display_fmt_ = {
            "state": "{}",
        }

    class state_duration(TdfReadingBase):
        """Duration an Infuse-IoT application state was asserted for"""

        ID = 57
        NAME = "STATE_DURATION"
        _fields_ = [
            ("state", ctypes.c_uint8),
            ("duration", ctypes.c_uint32),
        ]
        _pack_ = 1
        _postfix_ = {
            "state": "",
            "duration": "",
        }
        _display_fmt_ = {
            "state": "{}",
            "duration": "{}",
        }

    class pcm_16bit_chan_left(TdfReadingBase):
        """16bit PCM (Audio) data for the left channel"""

        ID = 58
        NAME = "PCM_16BIT_CHAN_LEFT"
        _fields_ = [
            ("val", ctypes.c_int16),
        ]
        _pack_ = 1
        _postfix_ = {
            "val": "",
        }
        _display_fmt_ = {
            "val": "{}",
        }

    class pcm_16bit_chan_right(TdfReadingBase):
        """16bit PCM (Audio) data for the right channel"""

        ID = 59
        NAME = "PCM_16BIT_CHAN_RIGHT"
        _fields_ = [
            ("val", ctypes.c_int16),
        ]
        _pack_ = 1
        _postfix_ = {
            "val": "",
        }
        _display_fmt_ = {
            "val": "{}",
        }

    class pcm_16bit_chan_dual(TdfReadingBase):
        """16bit PCM (Audio) data for both the left and right channels"""

        ID = 60
        NAME = "PCM_16BIT_CHAN_DUAL"
        _fields_ = [
            ("left", ctypes.c_int16),
            ("right", ctypes.c_int16),
        ]
        _pack_ = 1
        _postfix_ = {
            "left": "",
            "right": "",
        }
        _display_fmt_ = {
            "left": "{}",
            "right": "{}",
        }


id_type_mapping: dict[int, type[TdfReadingBase]] = {
    readings.announce.ID: readings.announce,
    readings.battery_state.ID: readings.battery_state,
    readings.ambient_temp_pres_hum.ID: readings.ambient_temp_pres_hum,
    readings.ambient_temperature.ID: readings.ambient_temperature,
    readings.time_sync.ID: readings.time_sync,
    readings.reboot_info.ID: readings.reboot_info,
    readings.announce_v2.ID: readings.announce_v2,
    readings.acc_2g.ID: readings.acc_2g,
    readings.acc_4g.ID: readings.acc_4g,
    readings.acc_8g.ID: readings.acc_8g,
    readings.acc_16g.ID: readings.acc_16g,
    readings.gyr_125dps.ID: readings.gyr_125dps,
    readings.gyr_250dps.ID: readings.gyr_250dps,
    readings.gyr_500dps.ID: readings.gyr_500dps,
    readings.gyr_1000dps.ID: readings.gyr_1000dps,
    readings.gyr_2000dps.ID: readings.gyr_2000dps,
    readings.gcs_wgs84_llha.ID: readings.gcs_wgs84_llha,
    readings.ubx_nav_pvt.ID: readings.ubx_nav_pvt,
    readings.lte_conn_status.ID: readings.lte_conn_status,
    readings.globalstar_pkt.ID: readings.globalstar_pkt,
    readings.acc_magnitude_std_dev.ID: readings.acc_magnitude_std_dev,
    readings.activity_metric.ID: readings.activity_metric,
    readings.algorithm_output.ID: readings.algorithm_output,
    readings.runtime_error.ID: readings.runtime_error,
    readings.charger_en_control.ID: readings.charger_en_control,
    readings.gnss_fix_info.ID: readings.gnss_fix_info,
    readings.bluetooth_connection.ID: readings.bluetooth_connection,
    readings.bluetooth_rssi.ID: readings.bluetooth_rssi,
    readings.bluetooth_data_throughput.ID: readings.bluetooth_data_throughput,
    readings.algorithm_class_histogram.ID: readings.algorithm_class_histogram,
    readings.algorithm_class_time_series.ID: readings.algorithm_class_time_series,
    readings.lte_tac_cells.ID: readings.lte_tac_cells,
    readings.wifi_ap_info.ID: readings.wifi_ap_info,
    readings.device_tilt.ID: readings.device_tilt,
    readings.nrf9x_gnss_pvt.ID: readings.nrf9x_gnss_pvt,
    readings.battery_charge_accumulated.ID: readings.battery_charge_accumulated,
    readings.infuse_bluetooth_rssi.ID: readings.infuse_bluetooth_rssi,
    readings.adc_raw_8.ID: readings.adc_raw_8,
    readings.adc_raw_16.ID: readings.adc_raw_16,
    readings.adc_raw_32.ID: readings.adc_raw_32,
    readings.annotation.ID: readings.annotation,
    readings.lora_rx.ID: readings.lora_rx,
    readings.lora_tx.ID: readings.lora_tx,
    readings.idx_array_freq.ID: readings.idx_array_freq,
    readings.idx_array_period.ID: readings.idx_array_period,
    readings.wifi_connected.ID: readings.wifi_connected,
    readings.wifi_connection_failed.ID: readings.wifi_connection_failed,
    readings.wifi_disconnected.ID: readings.wifi_disconnected,
    readings.network_scan_count.ID: readings.network_scan_count,
    readings.exception_stack_frame.ID: readings.exception_stack_frame,
    readings.battery_voltage.ID: readings.battery_voltage,
    readings.battery_soc.ID: readings.battery_soc,
    readings.state_event_set.ID: readings.state_event_set,
    readings.state_event_cleared.ID: readings.state_event_cleared,
    readings.state_duration.ID: readings.state_duration,
    readings.pcm_16bit_chan_left.ID: readings.pcm_16bit_chan_left,
    readings.pcm_16bit_chan_right.ID: readings.pcm_16bit_chan_right,
    readings.pcm_16bit_chan_dual.ID: readings.pcm_16bit_chan_dual,
}

__all__ = [
    "structs",
    "readings",
    "id_type_mapping",
]



from collections.abc import Mapping
from typing import TYPE_CHECKING, Any, TypeVar

from attrs import define as _attrs_define
from attrs import field as _attrs_field

if TYPE_CHECKING:
    from ..models.definitions_field_definition import DefinitionsFieldDefinition


T = TypeVar("T", bound="DefinitionsStructDefinition")


@_attrs_define
class DefinitionsStructDefinition:
    """
    Attributes:
        description (str):
        fields (list['DefinitionsFieldDefinition']):
    """

    description: str
    fields: list["DefinitionsFieldDefinition"]
    additional_properties: dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> dict[str, Any]:
        description = self.description

        fields = []
        for fields_item_data in self.fields:
            fields_item = fields_item_data.to_dict()
            fields.append(fields_item)

        field_dict: dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update(
            {
                "description": description,
                "fields": fields,
            }
        )

        return field_dict

    @classmethod
    def from_dict(cls: type[T], src_dict: Mapping[str, Any]) -> T:
        from ..models.definitions_field_definition import DefinitionsFieldDefinition

        d = dict(src_dict)
        description = d.pop("description")

        fields = []
        _fields = d.pop("fields")
        for fields_item_data in _fields:
            fields_item = DefinitionsFieldDefinition.from_dict(fields_item_data)

            fields.append(fields_item)

        definitions_struct_definition = cls(
            description=description,
            fields=fields,
        )

        definitions_struct_definition.additional_properties = d
        return definitions_struct_definition

    @property
    def additional_keys(self) -> list[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from enum import Enum


class DefinitionsRPCCommandDefaultAuth(str, Enum):
    EPACKET_AUTH_DEVICE = "EPACKET_AUTH_DEVICE"
    EPACKET_AUTH_NETWORK = "EPACKET_AUTH_NETWORK"

    def __str__(self) -> str:
        return str(self.value)



from collections.abc import Mapping
from typing import TYPE_CHECKING, Any, TypeVar, Union

from attrs import define as _attrs_define
from attrs import field as _attrs_field

from ..types import UNSET, Unset

if TYPE_CHECKING:
    from ..models.rpc_params import RPCParams
    from ..models.uplink_route import UplinkRoute


T = TypeVar("T", bound="RpcRsp")


@_attrs_define
class RpcRsp:
    """
    Attributes:
        route (UplinkRoute):
        return_code (int): Return code of RPC
        params (Union[Unset, RPCParams]): RPC request or response params (must be a JSON object with string or embedded
            json values - numbers sent as decimal strings) Example: {'primitive_vaue': '1000', 'struct_value': {'field':
            'value'}}.
        params_encoded (Union[Unset, str]): Base64 encoded params (provided if there was an issue decoding the RPC
            params)
    """

    route: "UplinkRoute"
    return_code: int
    params: Union[Unset, "RPCParams"] = UNSET
    params_encoded: Unset | str = UNSET
    additional_properties: dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> dict[str, Any]:
        route = self.route.to_dict()

        return_code = self.return_code

        params: Unset | dict[str, Any] = UNSET
        if not isinstance(self.params, Unset):
            params = self.params.to_dict()

        params_encoded = self.params_encoded

        field_dict: dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update(
            {
                "route": route,
                "returnCode": return_code,
            }
        )
        if params is not UNSET:
            field_dict["params"] = params
        if params_encoded is not UNSET:
            field_dict["paramsEncoded"] = params_encoded

        return field_dict

    @classmethod
    def from_dict(cls: type[T], src_dict: Mapping[str, Any]) -> T:
        from ..models.rpc_params import RPCParams
        from ..models.uplink_route import UplinkRoute

        d = dict(src_dict)
        route = UplinkRoute.from_dict(d.pop("route"))

        return_code = d.pop("returnCode")

        _params = d.pop("params", UNSET)
        params: Unset | RPCParams
        if isinstance(_params, Unset):
            params = UNSET
        else:
            params = RPCParams.from_dict(_params)

        params_encoded = d.pop("paramsEncoded", UNSET)

        rpc_rsp = cls(
            route=route,
            return_code=return_code,
            params=params,
            params_encoded=params_encoded,
        )

        rpc_rsp.additional_properties = d
        return rpc_rsp

    @property
    def additional_keys(self) -> list[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



#!/usr/bin/env python3

"""Serial to Bluetooth gateway control tool"""

__author__ = "Jordan Yates"
__copyright__ = "Copyright 2024, Embeint Holdings Pty Ltd"

import argparse
import base64
import ctypes
import io
import queue
import random
import sys
import threading
import time
from collections.abc import Callable

import cryptography
import cryptography.exceptions

import infuse_iot.definitions.rpc as defs
import infuse_iot.definitions.tdf as tdf_defs
import infuse_iot.epacket.interface as interface
from infuse_iot import rpc, tdf
from infuse_iot.commands import InfuseCommand
from infuse_iot.common import InfuseID, InfuseType
from infuse_iot.database import (
    DeviceDatabase,
    NoKeyError,
)
from infuse_iot.epacket.packet import (
    Auth,
    CtypeBtGattFrame,
    HopOutput,
    PacketOutputRouted,
    PacketReceived,
)
from infuse_iot.serial_comms import PyOcdPort, RttPort, SerialFrame, SerialLike, SerialPort
from infuse_iot.socket_comms import (
    ClientNotification,
    ClientNotificationConnectionCreated,
    ClientNotificationConnectionDropped,
    ClientNotificationConnectionFailed,
    ClientNotificationEpacketReceived,
    ClientNotificationObservedDevices,
    GatewayRequestConnectionRelease,
    GatewayRequestConnectionRequest,
    GatewayRequestEpacketSend,
    GatewayRequestObservedDevices,
    LocalServer,
    default_multicast_address,
)
from infuse_iot.util.argparse import ValidFile
from infuse_iot.util.console import Console
from infuse_iot.util.os import is_wsl
from infuse_iot.util.threading import SignaledThread


class LocalRpcServer:
    """Basic class supporting locally generated commands"""

    def __init__(self, database: DeviceDatabase):
        self._cnt = random.randint(0, 2**31)
        self._ddb = database
        self._queued: dict[int, Callable | None] = {}

    def generate(self, command: int, args: bytes, auth: Auth, cb: Callable | None):
        """Generate RPC packet from arguments"""
        cmd_bytes = bytes(rpc.RequestHeader(self._cnt, command)) + args
        cmd_pkt = PacketOutputRouted(
            [HopOutput.serial(auth)],
            InfuseType.RPC_CMD,
            cmd_bytes,
        )
        assert self._ddb.gateway is not None
        cmd_pkt.route[0].infuse_id = self._ddb.gateway
        self._queued[self._cnt] = cb
        self._cnt += 1
        return cmd_pkt

    def handle(self, pkt: PacketReceived):
        """Handle received packets"""
        # Only care about RPC responses
        if pkt.ptype != InfuseType.RPC_RSP:
            return

        # Inspect the response header
        header = rpc.ResponseHeader.from_buffer_copy(pkt.payload)

        # Was this a BT connect response with key information?
        if header.command_id == defs.bt_connect_infuse.COMMAND_ID:
            resp = defs.bt_connect_infuse.response.from_buffer_copy(pkt.payload[ctypes.sizeof(header) :])
            if_addr = interface.Address.BluetoothLeAddr.from_rpc_struct(resp.peer)
            infuse_id = self._ddb.infuse_id_from_bluetooth(if_addr)
            if infuse_id is None:
                Console.log_error(f"Infuse ID of {if_addr} not known")
            else:
                self._ddb.observe_security_state(
                    infuse_id,
                    bytes(resp.cloud_public_key),
                    bytes(resp.device_public_key),
                    resp.network_id,
                )

        # Determine if the response is to a command we initiated
        if header.request_id not in self._queued:
            return

        # Run the callback
        cb = self._queued.pop(header.request_id)
        if cb is not None:
            cb(pkt, header.return_code, pkt.payload[ctypes.sizeof(header) :])


class CommonThreadState:
    def __init__(
        self,
        server: LocalServer | None,
        port: SerialLike,
        ddb: DeviceDatabase,
        rpc_server: LocalRpcServer,
    ):
        self.server = server
        self.port = port
        self.ddb = ddb
        self.rpc = rpc_server

    def notification_broadcast(self, notification: ClientNotification):
        if self.server:
            self.server.broadcast(notification)

    def query_device_key(self, cb_event: threading.Event | None = None):
        def security_state_done(pkt: PacketReceived, _: int, response: bytes):
            cloud_key = response[:32]
            device_key = response[32:64]
            network_id = int.from_bytes(response[64:68], "little")

            self.ddb.observe_security_state(pkt.route[0].infuse_id, cloud_key, device_key, network_id)
            if cb_event is not None:
                cb_event.set()

        # Generate security_state RPC
        cmd_pkt = self.rpc.generate(30000, random.randbytes(16), Auth.NETWORK, security_state_done)
        encrypted = cmd_pkt.to_serial(self.ddb)
        # Write to serial port
        Console.log_tx(cmd_pkt.ptype, len(encrypted))
        self.port.write(encrypted)
        if cb_event is not None:
            # Wait for the response
            cb_event.wait(1.0)


class SerialRxThread(SignaledThread):
    """Receive serial frames from the serial port"""

    def __init__(self, common: CommonThreadState, log: io.TextIOWrapper):
        self._common = common
        self._reconstructor = SerialFrame.reconstructor()
        self._reconstructor.send(None)
        self._line = ""
        self._log = log
        self._next_ping = 0.0
        self._tdf_decoder = tdf.TDF()
        super().__init__(self._iter)

    def _iter(self) -> None:
        # Read bytes from serial port
        rx = self._common.port.read_bytes(1024)
        if len(rx) == 0:
            return
        for b in rx:
            frame_byte, frame = self._reconstructor.send(b)
            if frame and self._common.server is not None:
                self._handle_serial_frame(frame)

            if not frame_byte:
                c = chr(b)
                if c == "\n":
                    if self._log is not None:
                        self._log.write(self._line)
                    print(self._line)
                    self._line = ""
                else:
                    self._line += c

    def _handle_memfault_pkt(self, pkt: PacketReceived):
        class memfault_chunk_header(ctypes.LittleEndianStructure):
            _fields_ = [
                ("len", ctypes.c_uint16),
                ("cnt", ctypes.c_uint8),
            ]
            _pack_ = 1

        p = pkt.payload
        while len(p) > 0:
            hdr = memfault_chunk_header.from_buffer_copy(p)
            chunk = p[3 : 3 + hdr.len]
            p = p[3 + hdr.len :]
            print(f"Memfault Chunk {hdr.cnt:3d}: {base64.b64encode(chunk).decode('utf-8')}")

    def _handle_local_tdf(self, pkt: PacketReceived):
        if self._common.server is None:
            # No-one to broadcast events to
            return
        for reading in self._tdf_decoder.decode(pkt.payload):
            if isinstance(reading.data[0], tdf_defs.readings.bluetooth_connection) and reading.data[0].connected == 0:
                if_addr = interface.Address.BluetoothLeAddr.from_tdf_struct(reading.data[0].address)
                infuse_id = self._common.ddb.infuse_id_from_bluetooth(if_addr)
                if infuse_id:
                    self._common.notification_broadcast(ClientNotificationConnectionDropped(infuse_id))

    def _handle_serial_frame(self, frame: bytearray):
        try:
            # Decode the serial packet
            try:
                decoded = PacketReceived.from_serial(self._common.ddb, frame)
            except NoKeyError:
                assert self._common.ddb.gateway is not None
                if not self._common.ddb.has_network_id(self._common.ddb.gateway):
                    # Need to know network ID before we can query the device key
                    if time.time() >= self._next_ping:
                        self._next_ping = time.time() + 1.1
                        self._common.port.ping()
                        Console.log_info(f"Dropping {len(frame)} byte packet to query network ID...")
                    else:
                        Console.log_info(f"Dropping {len(frame)} byte packet...")
                else:
                    self._common.query_device_key(None)
                    Console.log_info(f"Dropping {len(frame)} byte packet to query device key...")
                return
            except cryptography.exceptions.InvalidTag as e:
                Console.log_error(f"Failed to decode {len(frame)} byte packet {e}")
                return

            # Iterate over all contained subpackets
            for pkt in decoded:
                Console.log_rx(pkt.ptype, len(frame))
                # Handle any local TDFs
                if len(pkt.route) == 1 and pkt.ptype == InfuseType.TDF:
                    self._handle_local_tdf(pkt)
                # Handle any local RPC responses
                self._common.rpc.handle(pkt)
                # Handle any Memfault chunks
                if pkt.ptype == InfuseType.MEMFAULT_CHUNK:
                    self._handle_memfault_pkt(pkt)
                # Proactively requery keys
                elif pkt.ptype == InfuseType.KEY_IDS:
                    self._common.query_device_key(None)

                # Forward to clients
                notification = ClientNotificationEpacketReceived(pkt)
                self._common.notification_broadcast(notification)
        except (ValueError, KeyError) as e:
            print(f"Decode failed ({e})")


class SerialTxThread(SignaledThread):
    """Send serial frames down the serial port"""

    def __init__(
        self,
        common: CommonThreadState,
    ):
        self._common = common
        self._connected: dict[int, int] = {}
        self._queue: queue.Queue = queue.Queue()
        super().__init__(self._iter)

    def send(self, pkt):
        """Queue packet for transmission"""
        self._queue.put(pkt)

    def _handle_epacket_send(self, req: GatewayRequestEpacketSend):
        if self._common.ddb.gateway is None:
            Console.log_error("Gateway address unknown")
            return

        pkt = req.epacket

        # Construct routed output
        if pkt.infuse_id == InfuseID.GATEWAY or pkt.infuse_id == self._common.ddb.gateway:
            routed = PacketOutputRouted(
                [HopOutput(self._common.ddb.gateway, interface.ID.SERIAL, pkt.auth)],
                pkt.ptype,
                pkt.payload,
            )
        else:
            gateway = self._common.ddb.gateway
            serial = HopOutput(gateway, interface.ID.SERIAL, Auth.DEVICE)
            bt = HopOutput(pkt.infuse_id, interface.ID.BT_CENTRAL, pkt.auth)
            routed = PacketOutputRouted(
                [serial, bt],
                pkt.ptype,
                pkt.payload,
            )

        # Do we have the device public keys we need?
        for hop in routed.route:
            if hop.auth == Auth.DEVICE and not self._common.ddb.has_public_key(hop.infuse_id):
                cb_event = threading.Event()
                self._common.query_device_key(cb_event)

        # Encode and encrypt payload
        encrypted = routed.to_serial(self._common.ddb)

        # Write to serial port
        Console.log_tx(routed.ptype, len(encrypted))
        self._common.port.write(encrypted)

    def _bt_connect_cb(self, pkt: PacketReceived, rc: int, response: bytes):
        resp = defs.bt_connect_infuse.response.from_buffer_copy(pkt.payload[ctypes.sizeof(rpc.ResponseHeader) :])
        if_addr = interface.Address.BluetoothLeAddr.from_rpc_struct(resp.peer)
        infuse_id = self._common.ddb.infuse_id_from_bluetooth(if_addr)

        assert infuse_id is not None, "ID was required to initiate connection?"
        assert self._common.server is not None

        rsp: ClientNotification
        if rc < 0:
            rsp = ClientNotificationConnectionFailed(infuse_id)
        else:
            if infuse_id in self._connected:
                self._connected[infuse_id] += 1
            else:
                self._connected[infuse_id] = 1
            rsp = ClientNotificationConnectionCreated(infuse_id, 244 - ctypes.sizeof(CtypeBtGattFrame) - 16)
        self._common.notification_broadcast(rsp)

    def _handle_conn_request(self, req: GatewayRequestConnectionRequest):
        assert self._common.server is not None

        if req.infuse_id == InfuseID.GATEWAY or req.infuse_id == self._common.ddb.gateway:
            # Local gateway always connected
            self._common.notification_broadcast(ClientNotificationConnectionCreated(req.infuse_id, 512))
            return

        state = self._common.ddb.devices.get(req.infuse_id, None)
        if state is None or state.bt_addr is None:
            self._common.notification_broadcast(ClientNotificationConnectionFailed(req.infuse_id))
            return

        subs = 0
        bt_char = defs.rpc_enum_infuse_bt_characteristic
        if req.data_types & req.DataType.COMMAND:
            subs |= bt_char.COMMAND
        if req.data_types & req.DataType.DATA:
            subs |= bt_char.DATA
        if req.data_types & req.DataType.LOGGING:
            subs |= bt_char.LOGGING

        # Multiple connection users, subscribe all
        if req.infuse_id in self._connected:
            subs = bt_char.COMMAND | bt_char.DATA | bt_char.LOGGING

        connect_args = defs.bt_connect_infuse.request(
            state.bt_addr.to_rpc_struct(),
            req.timeout_ms,
            subs,
            0,
        )
        cmd = self._common.rpc.generate(
            defs.bt_connect_infuse.COMMAND_ID,
            bytes(connect_args),
            Auth.DEVICE,
            self._bt_connect_cb,
        )
        encrypted = cmd.to_serial(self._common.ddb)
        Console.log_tx(cmd.ptype, len(encrypted))
        self._common.port.write(encrypted)

    def _handle_conn_release(self, req: GatewayRequestConnectionRelease):
        if req.infuse_id == InfuseID.GATEWAY or req.infuse_id == self._common.ddb.gateway:
            # Local gateway always connected
            return

        state = self._common.ddb.devices.get(req.infuse_id, None)
        if state is None or state.bt_addr is None:
            # Unknown device, nothing to do
            return

        if req.infuse_id in self._connected:
            # Decrement reference count
            self._connected[req.infuse_id] -= 1
            if self._connected[req.infuse_id] > 0:
                # Someone is still using the connection
                return
            self._connected.pop(req.infuse_id)

        disconnect_args = defs.bt_disconnect.request(state.bt_addr.to_rpc_struct())
        cmd = self._common.rpc.generate(defs.bt_disconnect.COMMAND_ID, bytes(disconnect_args), Auth.DEVICE, None)
        encrypted = cmd.to_serial(self._common.ddb)
        Console.log_tx(cmd.ptype, len(encrypted))
        self._common.port.write(encrypted)

    def _handle_observed_devices(self):
        if self._common.server is None:
            raise RuntimeError
        observed_devices = {}
        for device, state in self._common.ddb.devices.items():
            info = {}
            if state.network_id is not None:
                info["network_id"] = state.network_id
            if state.device_id is not None:
                info["device_id"] = state.device_id
            if self._common.ddb.gateway == device:
                info["gateway"] = True
            observed_devices[device] = info
        self._common.notification_broadcast(ClientNotificationObservedDevices(observed_devices))

    def _iter(self) -> None:
        if self._common.server is None:
            time.sleep(1.0)
            return

        # Loop while there are packets to send
        while req := self._common.server.receive():
            if isinstance(req, GatewayRequestEpacketSend):
                self._handle_epacket_send(req)
            elif isinstance(req, GatewayRequestConnectionRequest):
                self._handle_conn_request(req)
            elif isinstance(req, GatewayRequestConnectionRelease):
                self._handle_conn_release(req)
            elif isinstance(req, GatewayRequestObservedDevices):
                self._handle_observed_devices()
            else:
                Console.log_error(f"Unhandled request {type(req)}")


class SubCommand(InfuseCommand):
    NAME = "gateway"
    HELP = "Connect to a local gateway device"
    DESCRIPTION = "Connect to a gateway device over serial and route commands to Bluetooth devices"

    @classmethod
    def add_parser(cls, parser):
        # COM ports are not valid files
        serial_type = str if sys.platform == "win32" else ValidFile
        group = parser.add_mutually_exclusive_group(required=True)
        group.add_argument("--serial", type=serial_type, help="Gateway serial port")
        group.add_argument("--rtt", type=str, help="RTT serial port")
        group.add_argument("--pyocd", type=str, help="RTT via PyOCD")
        parser.add_argument(
            "--display-only",
            "-d",
            action="store_true",
            help="No networking, only display serial",
        )
        parser.add_argument(
            "--log",
            "-l",
            metavar="filename",
            const=f"{int(time.time())}_log.txt",
            nargs="?",
            type=argparse.FileType("w"),
            help="Save serial output to file",
        )
        parser.add_argument("--baud", type=int, default=115200, help="Baudrate for serial port")

    def __init__(self, args: argparse.Namespace):
        self.port: SerialLike
        if args.serial is not None:
            if is_wsl() and args.baud > 115200:
                Console.log_info("High baudrates can result in dropped data on WSL (from USB passthrough)")
            self.port = SerialPort(args.serial, args.baud)
        elif args.rtt is not None:
            self.port = RttPort(args.rtt)
        elif args.pyocd is not None:
            self.port = PyOcdPort(args.pyocd)
        self.ddb = DeviceDatabase()
        if args.display_only:
            self.server = None
        else:
            self.server = LocalServer(default_multicast_address())
        self.rpc_server = LocalRpcServer(self.ddb)
        self._common = CommonThreadState(self.server, self.port, self.ddb, self.rpc_server)
        self.log = args.log
        Console.init()

    def run(self):
        # Open the serial port
        self.port.open()
        Console.log_info(f"Port '{str(self.port)}' opened")
        # Ping the port to get the local device ID
        self.port.ping()

        # Start threads
        rx_thread = SerialRxThread(self._common, self.log)
        tx_thread = SerialTxThread(self._common)
        rx_thread.start()
        tx_thread.start()

        # Run until 'Ctrl+C' or a thread dies
        try:
            while rx_thread.is_alive() and tx_thread.is_alive():
                time.sleep(0.5)
        except KeyboardInterrupt:
            pass
        finally:
            rx_thread.stop()
            tx_thread.stop()

        # Wait for threads to terminate
        rx_thread.join(1.0)
        tx_thread.join(1.0)

        # Cleanup serial port
        try:
            self.port.close()
        except Exception:
            pass

        if self.log:
            self.log.flush()



#!/usr/bin/env python3

import binascii

from rich.progress import (
    DownloadColumn,
    Progress,
    TransferSpeedColumn,
)

import infuse_iot.definitions.rpc as defs
from infuse_iot.commands import Auth, InfuseRpcCommand
from infuse_iot.definitions.rpc import rpc_enum_file_action
from infuse_iot.zephyr.errno import errno


class file_write_basic(InfuseRpcCommand, defs.file_write_basic):
    RPC_DATA_SEND = True

    @classmethod
    def add_parser(cls, parser):
        parser.add_argument(
            "--file",
            "-f",
            type=str,
            required=True,
            help="File to write",
        )
        group = parser.add_mutually_exclusive_group(required=True)
        group.add_argument(
            "--discard",
            dest="action",
            action="store_const",
            const=rpc_enum_file_action.DISCARD,
            help="Discard data without action",
        )
        group.add_argument(
            "--dfu",
            dest="action",
            action="store_const",
            const=rpc_enum_file_action.APP_IMG,
            help="Write complete image file and perform DFU",
        )
        group.add_argument(
            "--cpatch",
            dest="action",
            action="store_const",
            const=rpc_enum_file_action.APP_CPATCH,
            help="Write diff image file and perform DFU",
        )
        group.add_argument(
            "--bt-ctlr-dfu",
            dest="action",
            action="store_const",
            const=rpc_enum_file_action.BT_CTLR_IMG,
            help="Write Bluetooth controller image file and perform DFU",
        )
        group.add_argument(
            "--bt-ctlr-cpatch",
            dest="action",
            action="store_const",
            const=rpc_enum_file_action.BT_CTLR_CPATCH,
            help="Write Bluetooth controller diff file and perform DFU",
        )
        group.add_argument(
            "--nrf91-modem",
            dest="action",
            action="store_const",
            const=rpc_enum_file_action.NRF91_MODEM_DIFF,
            help="nRF91 LTE modem diff upgrade",
        )
        group.add_argument(
            "--for-copy",
            dest="action",
            action="store_const",
            const=rpc_enum_file_action.FILE_FOR_COPY,
            help="File to copy to other device",
        )

    def __init__(self, args):
        self.file = args.file
        self.action = args.action
        self.progress = Progress(
            *Progress.get_default_columns(),
            DownloadColumn(),
            TransferSpeedColumn(),
        )
        self.task = None

        with open(self.file, "rb") as f:
            self.payload = f.read()

    def auth_level(self):
        return Auth.NETWORK

    def request_struct(self):
        return self.request(self.action, binascii.crc32(self.payload))

    def data_payload(self):
        print("Preparing for file upload...")
        return self.payload

    def data_progress_cb(self, offset):
        if self.task is None:
            self.progress.start()
            self.task = self.progress.add_task("Writing...", total=len(self.payload))
        self.progress.update(self.task, completed=offset)

    def handle_response(self, return_code, response):
        self.progress.stop()

        if return_code != 0:
            print(f"Failed to write file ({errno.strerror(-return_code)})")
            return
        print("File written")
        print(f"\tLength: {response.recv_len}")
        print(f"\t   CRC: 0x{response.recv_crc:08x}")



#!/usr/bin/env python3

import ctypes

from elftools.dwarf.die import DIE
from elftools.dwarf.dwarf_expr import DW_OP_name2opcode
from elftools.dwarf.dwarfinfo import DWARFInfo
from elftools.elf.elffile import ELFFile
from elftools.elf.sections import Symbol, SymbolTableSection
from typing_extensions import Self

TYPEDEF_TO_CTYPE: dict[str, type[ctypes._SimpleCData]] = {
    "int8_t": ctypes.c_int8,
    "uint8_t": ctypes.c_uint8,
    "int16_t": ctypes.c_int16,
    "uint16_t": ctypes.c_uint16,
    "int32_t": ctypes.c_int32,
    "uint32_t": ctypes.c_uint32,
    "int64_t": ctypes.c_int64,
    "uint64_t": ctypes.c_uint64,
}

DWARF_TO_CTYPE: dict[str, type[ctypes._SimpleCData]] = {
    "_Bool": ctypes.c_bool,
    "unsigned char": ctypes.c_ubyte,
    "signed char": ctypes.c_byte,
    "char": ctypes.c_byte,
    "short unsigned int": ctypes.c_uint16,
    "unsigned int": ctypes.c_uint32,
    "int": ctypes.c_int32,
    "long int": ctypes.c_int32,
    "long long int": ctypes.c_int64,
    "long long unsigned int": ctypes.c_uint64,
    "float": ctypes.c_float,
    "double": ctypes.c_double,
}


def symbols_from_name(elf: ELFFile, name: str) -> list[Symbol]:
    """Get a list of symbols from an ELF file with names matching the provided string"""
    symtab = None

    # Locate the symbol table
    for section in elf.iter_sections():
        if isinstance(section, SymbolTableSection):
            symtab = section
            break

    if not symtab:
        return []

    symbols = []
    # Search for the symbol in the symbol table
    for symbol in symtab.iter_symbols():
        if symbol.name == name:
            symbols.append(symbol)
    return symbols


def symbol_from_address(elf: ELFFile, address: int) -> Symbol | None:
    """Get a list of symbols from an ELF file with names matching the provided string"""
    symtab = None

    # Locate the symbol table
    for section in elf.iter_sections():
        if isinstance(section, SymbolTableSection):
            symtab = section
            break

    if not symtab:
        return None

    # Search for the symbol in the symbol table
    for symbol in symtab.iter_symbols():
        start = symbol.entry["st_value"]
        size = symbol.entry["st_size"]
        end = start + size - 1
        if start <= address <= end:
            return symbol
    return None


def dwarf_die_from_symbol(elf: ELFFile, symbol: Symbol) -> DIE | None:
    """Get a Debug Information Entry associated with a symbol (Global variables only)"""
    dwarfinfo = elf.get_dwarf_info()

    for CU in dwarfinfo.iter_CUs():
        for die in CU.iter_DIEs():
            if die.tag == "DW_TAG_variable" and "DW_AT_name" in die.attributes and "DW_AT_location" in die.attributes:
                die_name = die.attributes["DW_AT_name"].value.decode("utf-8")
                die_location = die.attributes.get("DW_AT_location").value
                # Not our symbol
                if die_name != symbol.name:
                    continue
                # Constant addresses are in a list of form [0x03, addr_bytes]
                if not isinstance(die_location, list):
                    continue
                if die_location[0] != DW_OP_name2opcode["DW_OP_addr"]:
                    continue
                address = int.from_bytes(die_location[1:], "little")
                if address == symbol.entry["st_value"]:
                    return die
    return None


def dwarf_die_file_info(elf: ELFFile, die: DIE) -> tuple[str | None, int]:
    file_attr = die.attributes["DW_AT_decl_file"]
    line_attr = die.attributes["DW_AT_decl_line"]

    dwarfinfo = elf.get_dwarf_info()
    lineprogram = dwarfinfo.line_program_for_CU(die.cu)
    if lineprogram is None:
        cu_filename = None
    else:
        cu_filename = lineprogram["file_entry"][file_attr.value - 1].name.decode("latin-1")

    return cu_filename, line_attr.value


class dwarf_field:
    def __init__(
        self,
        name: str,
        tag: str,
        ctype: type[ctypes._SimpleCData] | None,
        children: list[Self] | None,
        offset: int,
    ):
        self.name = name
        self.tag = tag
        self.ctype = ctype
        if children is None:
            self.children = []
        else:
            self.children = children
        self.offset = offset


def _type_from_dwarf_info(dwarfinfo: DWARFInfo, die: DIE):
    refaddr = die.attributes["DW_AT_type"].value + die.cu.cu_offset
    return dwarfinfo.get_DIE_from_refaddr(refaddr, die.cu)


def dwarf_die_variable_inf(
    dwarfinfo: DWARFInfo, die: DIE, offset: int = 0, name_override: str | None = None
) -> dwarf_field:
    type_die = _type_from_dwarf_info(dwarfinfo, die)

    if "DW_AT_name" in die.attributes:
        field_name = die.attributes["DW_AT_name"].value.decode("utf-8")
    else:
        field_name = "<unknown>"
    if "DW_AT_name" in type_die.attributes:
        type_name = type_die.attributes["DW_AT_name"].value.decode("utf-8")
    else:
        type_name = None
    info_name = name_override if name_override is not None else field_name

    if type_die.tag == "DW_TAG_array_type":
        count = 0
        for child in type_die.iter_children():
            if child.tag == "DW_TAG_subrange_type" and "DW_AT_upper_bound" in child.attributes:
                count = child.attributes["DW_AT_upper_bound"].value + 1
        element_die = _type_from_dwarf_info(dwarfinfo, type_die)

        children = []
        element_offset = 0
        for idx in range(count):
            child = dwarf_die_variable_inf(dwarfinfo, type_die, offset + element_offset, f"{info_name}[{idx}]")
            if "DW_AT_byte_size" in element_die.attributes:
                element_offset += element_die.attributes["DW_AT_byte_size"].value
            else:
                if child.ctype is not None:
                    element_offset += ctypes.sizeof(child.ctype)
            children.append(child)

        return dwarf_field(info_name, type_die.tag, None, children, offset)
    elif type_die.tag == "DW_TAG_structure_type":
        children = []
        for child in type_die.iter_children():
            field_offset = child.attributes["DW_AT_data_member_location"].value
            child_field = dwarf_die_variable_inf(dwarfinfo, child, offset + field_offset)
            children.append(child_field)

        return dwarf_field(info_name, type_die.tag, None, children, offset)

    elif type_die.tag == "DW_TAG_union_type":
        children = []
        for child in type_die.iter_children():
            child_field = dwarf_die_variable_inf(dwarfinfo, child, offset)
            children.append(child_field)
        return dwarf_field(info_name, type_die.tag, None, children, offset)
    elif type_die.tag == "DW_TAG_typedef":
        if type_name in TYPEDEF_TO_CTYPE:
            return dwarf_field(info_name, type_die.tag, TYPEDEF_TO_CTYPE[type_name], None, offset)
        else:
            return dwarf_die_variable_inf(dwarfinfo, type_die, offset, name_override)
    elif type_die.tag == "DW_TAG_base_type":
        if type_name not in DWARF_TO_CTYPE:
            raise NotImplementedError(f"'{type_name}' not known in DWARF_TO_CTYPE")
        return dwarf_field(info_name, type_die.tag, DWARF_TO_CTYPE[type_name], None, offset)
    elif type_die.tag == "DW_TAG_pointer_type":
        return dwarf_field(info_name, type_die.tag, ctypes.c_uint32, None, offset)
    elif type_die.tag in ["DW_TAG_enumeration_type", "DW_TAG_const_type"]:
        return dwarf_die_variable_inf(dwarfinfo, type_die, offset)
    else:
        raise NotImplementedError(type_die.tag)



#!/usr/bin/env python3

"""Plot CSV data"""

__author__ = "Jordan Yates"
__copyright__ = "Copyright 2024, Embeint Holdings Pty Ltd"

from infuse_iot.commands import InfuseCommand
from infuse_iot.util.argparse import ValidFile


class SubCommand(InfuseCommand):
    NAME = "csv_plot"
    HELP = "Plot CSV data"
    DESCRIPTION = "Plot CSV data"

    @classmethod
    def add_parser(cls, parser):
        parser.add_argument("--file", "-f", required=True, type=ValidFile)
        parser.add_argument("--start", type=str, default="2024-01-01", help="Display data after")

    def __init__(self, args):
        self.file = args.file
        self.start = args.start

    def run(self):
        import pandas as pd
        import plotly.express as px
        from dash import Dash, dcc, html

        df = pd.read_csv(self.file)

        mask = df["time"] >= self.start
        filtered_df = df.loc[mask]

        fig = px.line(
            filtered_df,
            x="time",
            y=filtered_df.columns.values[1:],
            title=str(self.file),
        )

        app = Dash()
        app.layout = html.Div([dcc.Graph(figure=fig)])

        app.run(debug=True)



from collections.abc import Mapping
from typing import TYPE_CHECKING, Any, TypeVar

from attrs import define as _attrs_define
from attrs import field as _attrs_field

if TYPE_CHECKING:
    from ..models.uplink_route import UplinkRoute


T = TypeVar("T", bound="ForwardedUplinkRoute")


@_attrs_define
class ForwardedUplinkRoute:
    """
    Attributes:
        device_id (str): The ID of the forwarding device as a hex string Example: d291d4d66bf0a955.
        route (UplinkRoute):
        auth (int): Auth level for the forwarding device packet
        rssi (int): RSSI of the packet
    """

    device_id: str
    route: "UplinkRoute"
    auth: int
    rssi: int
    additional_properties: dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> dict[str, Any]:
        device_id = self.device_id

        route = self.route.to_dict()

        auth = self.auth

        rssi = self.rssi

        field_dict: dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update(
            {
                "deviceId": device_id,
                "route": route,
                "auth": auth,
                "rssi": rssi,
            }
        )

        return field_dict

    @classmethod
    def from_dict(cls: type[T], src_dict: Mapping[str, Any]) -> T:
        from ..models.uplink_route import UplinkRoute

        d = dict(src_dict)
        device_id = d.pop("deviceId")

        route = UplinkRoute.from_dict(d.pop("route"))

        auth = d.pop("auth")

        rssi = d.pop("rssi")

        forwarded_uplink_route = cls(
            device_id=device_id,
            route=route,
            auth=auth,
            rssi=rssi,
        )

        forwarded_uplink_route.additional_properties = d
        return forwarded_uplink_route

    @property
    def additional_keys(self) -> list[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



#!/usr/bin/env python3

import os
import subprocess

import pytest

import infuse_iot.credentials as cred

assert "TOXTEMPDIR" in os.environ, "you must run these tests using tox"


def test_credentials():
    # Validate the credentials API

    try:
        cred.delete_api_key()
    except Exception as _:
        pass

    with pytest.raises(FileNotFoundError):
        cred.get_api_key()

    test_api_key = "ABCDEFGHIJKLMNOP"
    test_api_key_2 = "ABCDEFGHIJKLMNOP123456"

    output = subprocess.check_output(["infuse", "credentials", "--api-key-print"]).decode()
    assert "API Key: N/A" in output

    subprocess.check_output(["infuse", "credentials", "--api-key", test_api_key]).decode()
    assert test_api_key == cred.get_api_key()

    output = subprocess.check_output(["infuse", "credentials", "--api-key-print"]).decode()
    assert test_api_key in output

    cred.set_api_key(test_api_key_2)

    output = subprocess.check_output(["infuse", "credentials", "--api-key-print"]).decode()
    assert test_api_key_2 in output

    cred.delete_api_key()

    output = subprocess.check_output(["infuse", "credentials", "--api-key-print"]).decode()
    assert "API Key: N/A" in output



#!/usr/bin/env python3

# import ctypes
import argparse

from infuse_iot.definitions.kv import slots as kv_slots
from infuse_iot.definitions.kv import structs as kv_structs
from infuse_iot.zephyr import lte as z_lte
from infuse_iot.zephyr.errno import errno

from . import kv_read, lte_pdp_ctx


class lte_modem_info(kv_read.kv_read):
    HELP = "Get LTE modem information"
    DESCRIPTION = "Get LTE modem information"

    @classmethod
    def add_parser(cls, parser):
        return

    def __init__(self, _args):
        super().__init__(argparse.Namespace(keys=[40, 41, 42, 43, 44, 45, 46]))

    def handle_response(self, return_code, response):
        if return_code != 0:
            print(f"Failed to query modem info ({errno.strerror(-return_code)})")
            return

        def struct_decode(t, r):
            if r.len <= 0:
                return None
            return t.vla_from_buffer_copy(bytes(r.data))

        def str_decode(r):
            if r.len <= 0:
                return "Unknown"
            return str(kv_structs.kv_string.vla_from_buffer_copy(bytes(r.data)))

        modem_imei = struct_decode(kv_slots.lte_modem_imei, response.values[3])
        pdp_ctx = struct_decode(kv_slots.lte_pdp_config, response.values[5])
        system_modes = struct_decode(kv_slots.lte_networking_modes, response.values[6])

        if pdp_ctx:
            pdp_str = f'"{str(pdp_ctx.apn)}" ({lte_pdp_ctx.lte_pdp_ctx.PDPFamily(pdp_ctx.family).name})'
        else:
            pdp_str = "unknown"
        if system_modes:
            system_mode = z_lte.LteSystemMode(system_modes.modes)
            system_pref = z_lte.LteSystemPreference(system_modes.prefer)
            modes_str = f"{system_mode} (Prefer: {system_pref})"
        else:
            modes_str = "default"

        print(f"\t   Model: {str_decode(response.values[0])}")
        print(f"\tFirmware: {str_decode(response.values[1])}")
        print(f"\t     ESN: {str_decode(response.values[2])}")
        print(f"\t    IMEI: {modem_imei.imei}")
        print(f"\t     SIM: {str_decode(response.values[4])}")
        print(f"\t     APN: {pdp_str}")
        print(f"\t    Mode: {modes_str}")



#!/usr/bin/env python3

import copy
import ctypes
import enum
import time
from collections.abc import Generator

from infuse_iot.definitions import tdf as tdf_defs
from infuse_iot.generated import tdf_base
from infuse_iot.time import InfuseTime


def unknown_tdf_factory(tdf_id: int, tdf_len: int) -> type[tdf_base.TdfReadingBase]:
    class UnknownTDF(tdf_base.TdfReadingBase):
        NAME = str(tdf_id)
        ID = tdf_id
        _fields_ = [
            ("data", tdf_len * ctypes.c_uint8),
        ]
        _pack_ = 1
        _postfix_ = {"data": ""}
        _display_fmt_ = {"data": "{}"}

    return UnknownTDF


class TDF:
    class flags(enum.IntEnum):
        TIMESTAMP_NONE = 0x0000
        TIMESTAMP_ABSOLUTE = 0x4000
        TIMESTAMP_RELATIVE = 0x8000
        TIMESTAMP_EXTENDED_RELATIVE = 0xC000
        TIMESTAMP_MASK = 0xC000
        TIME_ARRAY = 0x1000
        DIFF_ARRAY = 0x2000
        IDX_ARRAY = 0x3000
        ARRAY_MASK = 0x3000
        ID_MASK = 0x0FFF

    class DiffType(enum.IntEnum):
        DIFF_16_8 = 1
        DIFF_32_8 = 2
        DIFF_32_16 = 3

    class CoreHeader(ctypes.LittleEndianStructure):
        _fields_ = [
            ("id_flags", ctypes.c_uint16),
            ("len", ctypes.c_uint8),
        ]
        _pack_ = 1

    class ArrayHeader(ctypes.LittleEndianStructure):
        _fields_ = [
            ("num", ctypes.c_uint8),
            ("period", ctypes.c_uint16),
        ]
        _pack_ = 1

    class AbsoluteTime(ctypes.LittleEndianStructure):
        _fields_ = [
            ("seconds", ctypes.c_uint32),
            ("subseconds", ctypes.c_uint16),
        ]
        _pack_ = 1

    class RelativeTime(ctypes.LittleEndianStructure):
        _fields_ = [
            ("offset", ctypes.c_uint16),
        ]
        _pack_ = 1

    class ExtendedRelativeTime(ctypes.LittleEndianStructure):
        _fields_ = [
            ("_offset", ctypes.c_uint8 * 3),
        ]
        _pack_ = 1

        @property
        def offset(self):
            return int.from_bytes(self._offset, byteorder="little", signed=True)

    class Reading:
        def __init__(
            self,
            tdf_id: int,
            reading_time: None | float,
            period: None | float,
            base_idx: None | int,
            data: list[tdf_base.TdfReadingBase],
        ):
            self.id = tdf_id
            self.time = reading_time
            self.period = period
            self.base_idx = base_idx
            self.data = data

    def __init__(self):
        pass

    @staticmethod
    def _buffer_pull(buffer: bytes, ctype: type[ctypes.LittleEndianStructure]):
        v = ctype.from_buffer_copy(buffer)
        b = buffer[ctypes.sizeof(ctype) :]
        return v, b

    @classmethod
    def _diff_expand(cls, buffer: bytes, tdf_len: int, diff_type: DiffType, diff_num: int) -> tuple[int, bytes]:
        t_in: type[ctypes._SimpleCData]
        t_diff: type[ctypes._SimpleCData]
        if diff_type == cls.DiffType.DIFF_16_8:
            t_in = ctypes.c_uint16
            t_diff = ctypes.c_int8
        elif diff_type == cls.DiffType.DIFF_32_8:
            t_in = ctypes.c_uint32
            t_diff = ctypes.c_int8
        elif diff_type == cls.DiffType.DIFF_32_16:
            t_in = ctypes.c_uint32
            t_diff = ctypes.c_int16
        else:
            raise RuntimeError(f"Unknown diff type {diff_type}")
        num_fields = tdf_len // ctypes.sizeof(t_in)

        class _tdf(ctypes.LittleEndianStructure):
            _fields_ = [("data", num_fields * t_in)]
            _pack_ = 1

        class _diff(ctypes.LittleEndianStructure):
            _fields_ = [("data", num_fields * t_diff)]
            _pack_ = 1

        class _complete(ctypes.LittleEndianStructure):
            _fields_ = [("base", _tdf), ("diffs", diff_num * _diff)]
            _pack_ = 1

        raw = _complete.from_buffer_copy(buffer)
        out: list[ctypes.LittleEndianStructure] = [_tdf.from_buffer_copy(buffer)]
        for idx in range(diff_num):
            next = copy.copy(out[-1])
            for f in range(num_fields):
                next.data[f] += raw.diffs[idx].data[f]
            out.append(next)

        expanded = b"".join([bytes(b) for b in out])
        return ctypes.sizeof(raw), expanded

    def decode(self, buffer: bytes, no_defs: bool = False) -> Generator[Reading, None, None]:
        buffer_time = None

        while len(buffer) > 3:
            header, buffer = self._buffer_pull(buffer, self.CoreHeader)
            time_flags = header.id_flags & self.flags.TIMESTAMP_MASK

            if header.id_flags in [0x0000, 0xFFFF]:
                break

            tdf_id = header.id_flags & 0x0FFF
            if no_defs:
                id_type = unknown_tdf_factory(tdf_id, header.len)
            else:
                try:
                    id_type = tdf_defs.id_type_mapping[tdf_id]
                except KeyError:
                    id_type = unknown_tdf_factory(tdf_id, header.len)

            if time_flags == self.flags.TIMESTAMP_NONE:
                reading_time = None
            elif time_flags == self.flags.TIMESTAMP_ABSOLUTE:
                t, buffer = self._buffer_pull(buffer, self.AbsoluteTime)
                buffer_time = t.seconds * 65536 + t.subseconds
                reading_time = InfuseTime.unix_time_from_epoch(buffer_time)
            elif time_flags == self.flags.TIMESTAMP_RELATIVE:
                t, buffer = self._buffer_pull(buffer, self.RelativeTime)
                buffer_time += t.offset
                reading_time = InfuseTime.unix_time_from_epoch(buffer_time)
            elif time_flags == self.flags.TIMESTAMP_EXTENDED_RELATIVE:
                t, buffer = self._buffer_pull(buffer, self.ExtendedRelativeTime)
                buffer_time += t.offset
                reading_time = InfuseTime.unix_time_from_epoch(buffer_time)
            else:
                raise RuntimeError("Unreachable time option")

            array_header = None
            base_idx = None
            array_type = header.id_flags & self.flags.ARRAY_MASK
            if array_type == self.flags.DIFF_ARRAY:
                array_header, buffer = self._buffer_pull(buffer, self.ArrayHeader)
                diff_type = array_header.num >> 6
                diff_num = array_header.num & 0x3F

                total_len, expanded = self._diff_expand(buffer, header.len, self.DiffType(diff_type), diff_num)
                buffer = buffer[total_len:]
                if buffer_time is None:
                    t_now = int(time.time() * 65536)
                    reading_time = InfuseTime.unix_time_from_epoch(t_now)
                else:
                    reading_time = InfuseTime.unix_time_from_epoch(buffer_time)
                data = [
                    id_type.from_buffer_consume(expanded[x : x + header.len]) for x in range(0, total_len, header.len)
                ]
            elif array_type == self.flags.TIME_ARRAY:
                array_header, buffer = self._buffer_pull(buffer, self.ArrayHeader)
                total_len = array_header.num * header.len
                total_data = buffer[:total_len]
                buffer = buffer[total_len:]

                if buffer_time is None:
                    t_now = int(time.time() * 65536)
                    reading_time = InfuseTime.unix_time_from_epoch(t_now)
                else:
                    reading_time = InfuseTime.unix_time_from_epoch(buffer_time)
                data = [
                    id_type.from_buffer_consume(total_data[x : x + header.len]) for x in range(0, total_len, header.len)
                ]
            elif array_type == self.flags.IDX_ARRAY:
                array_header, buffer = self._buffer_pull(buffer, self.ArrayHeader)
                total_len = array_header.num * header.len
                total_data = buffer[:total_len]
                buffer = buffer[total_len:]

                if time_flags != self.flags.TIMESTAMP_NONE:
                    assert buffer_time is not None
                    reading_time = InfuseTime.unix_time_from_epoch(buffer_time)
                base_idx = array_header.period
                data = [
                    id_type.from_buffer_consume(total_data[x : x + header.len]) for x in range(0, total_len, header.len)
                ]
            else:
                data_bytes = buffer[: header.len]
                buffer = buffer[header.len :]

                data = [id_type.from_buffer_consume(data_bytes)]

            period = None
            if array_header is not None and base_idx is None:
                period = array_header.period / 65536

            yield self.Reading(tdf_id, reading_time, period, base_idx, data)


if __name__ == "__main__":
    import sys

    if len(sys.argv) != 2:
        sys.exit("Expected `python -m infuse_iot.tdf /path/to/tdf.bin`")

    decoder = TDF()

    with open(sys.argv[1], "rb") as f:
        tdf_binary_blob = f.read(-1)

        # Check data is in the form we expect (single block, 2 byte prefix)
        if tdf_binary_blob[1] != 0x02:
            sys.exit("Expected second byte to be 0x02 (INFUSE_TDF)")

        try:
            for tdf in decoder.decode(tdf_binary_blob[2:]):
                print(f"TDF {tdf.id} @ t={tdf.time}: {tdf.data}")
        except Exception:
            pass



#!/usr/bin/env python3

import infuse_iot.generated.kv_definitions as _kv_base
from infuse_iot.util.internal import extension_load as _loader

if extension_module := _loader("kv_definitions"):

    class slots(_kv_base.slots, extension_module.slots):  # type: ignore
        pass

    class structs(_kv_base.structs, extension_module.structs):  # type: ignore
        pass

else:
    slots = _kv_base.slots  # type: ignore
    structs = _kv_base.structs  # type: ignore

__all__ = [
    "slots",
    "structs",
]



from collections.abc import Mapping
from typing import TYPE_CHECKING, Any, TypeVar

from attrs import define as _attrs_define
from attrs import field as _attrs_field

from ..types import UNSET, Unset

if TYPE_CHECKING:
    from ..models.new_rpc_req import NewRPCReq


T = TypeVar("T", bound="NewRPCMessage")


@_attrs_define
class NewRPCMessage:
    """
    Attributes:
        device_id (str): The ID of the device to send the RPC to as a hex string Example: d291d4d66bf0a955.
        rpc (NewRPCReq):
        send_wait_timeout_ms (Union[Unset, int]): Maximum time to wait (in milliseconds) for the device to send a
            packet. If 0 or not set, the RPC is sent immediately using the device's last route. Default: 60000.
    """

    device_id: str
    rpc: "NewRPCReq"
    send_wait_timeout_ms: Unset | int = 60000
    additional_properties: dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> dict[str, Any]:
        device_id = self.device_id

        rpc = self.rpc.to_dict()

        send_wait_timeout_ms = self.send_wait_timeout_ms

        field_dict: dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update(
            {
                "deviceId": device_id,
                "rpc": rpc,
            }
        )
        if send_wait_timeout_ms is not UNSET:
            field_dict["sendWaitTimeoutMs"] = send_wait_timeout_ms

        return field_dict

    @classmethod
    def from_dict(cls: type[T], src_dict: Mapping[str, Any]) -> T:
        from ..models.new_rpc_req import NewRPCReq

        d = dict(src_dict)
        device_id = d.pop("deviceId")

        rpc = NewRPCReq.from_dict(d.pop("rpc"))

        send_wait_timeout_ms = d.pop("sendWaitTimeoutMs", UNSET)

        new_rpc_message = cls(
            device_id=device_id,
            rpc=rpc,
            send_wait_timeout_ms=send_wait_timeout_ms,
        )

        new_rpc_message.additional_properties = d
        return new_rpc_message

    @property
    def additional_keys(self) -> list[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



#!/usr/bin/env python3


import infuse_iot.definitions.rpc as defs
from infuse_iot.commands import InfuseRpcCommand
from infuse_iot.definitions.rpc import (
    rpc_enum_bt_le_addr_type,
    rpc_struct_bt_addr_le,
)
from infuse_iot.util.argparse import BtLeAddress
from infuse_iot.util.ctypes import bytes_to_uint8
from infuse_iot.zephyr.errno import errno
from infuse_iot.zephyr.hci import error


class bt_mcumgr_reboot(InfuseRpcCommand, defs.bt_mcumgr_reboot):
    @classmethod
    def add_parser(cls, parser):
        parser.add_argument("--timeout", type=int, default=5000, help="Connection timeout (ms)")
        addr_group = parser.add_mutually_exclusive_group(required=True)
        addr_group.add_argument("--public", type=BtLeAddress, help="Public Bluetooth address")
        addr_group.add_argument("--random", type=BtLeAddress, help="Random Bluetooth address")

    def __init__(self, args):
        self.args = args

    def request_struct(self) -> defs.bt_mcumgr_reboot.request:
        if self.args.public:
            peer = rpc_struct_bt_addr_le(
                rpc_enum_bt_le_addr_type.PUBLIC,
                bytes_to_uint8(self.args.public.to_bytes(6, "little")),
            )
        else:
            peer = rpc_struct_bt_addr_le(
                rpc_enum_bt_le_addr_type.RANDOM,
                bytes_to_uint8(self.args.random.to_bytes(6, "little")),
            )

        return self.request(
            peer,
            self.args.timeout,
        )

    def handle_response(self, return_code, response):
        if return_code < 0:
            print(f"Failed to reboot ({errno.strerror(-return_code)})")
            return
        elif return_code > 0:
            print(f"Failed to reboot ({error.strerror(return_code)})")

        print("Reboot request sent")



#!/usr/bin/env python3

import enum
import json
import socket
import struct
import sys
from collections.abc import Generator
from contextlib import contextmanager
from typing import cast

from typing_extensions import Self

from infuse_iot.common import InfuseType
from infuse_iot.epacket.packet import HopReceived, PacketOutput, PacketReceived
from infuse_iot.generated.tdf_definitions import readings
from infuse_iot.tdf import TDF


def default_multicast_address():
    return ("224.1.1.1", 8751)


class ClientNotification:
    class Type(enum.IntEnum):
        EPACKET_RECV = 0
        CONNECTION_FAILED = 1
        CONNECTION_CREATED = 2
        CONNECTION_DROPPED = 3
        KNOWN_DEVICES = 4

    def to_json(self) -> dict:
        """Convert class to json dictionary"""
        raise NotImplementedError

    @classmethod
    def from_json(cls, values: dict) -> Self:
        """Reconstruct class from json dictionary"""

        if values["type"] == cls.Type.EPACKET_RECV:
            return cast(Self, ClientNotificationEpacketReceived.from_json(values))
        elif values["type"] == cls.Type.CONNECTION_FAILED:
            return cast(Self, ClientNotificationConnectionFailed.from_json(values))
        elif values["type"] == cls.Type.CONNECTION_CREATED:
            return cast(Self, ClientNotificationConnectionCreated.from_json(values))
        elif values["type"] == cls.Type.CONNECTION_DROPPED:
            return cast(Self, ClientNotificationConnectionDropped.from_json(values))
        elif values["type"] == cls.Type.KNOWN_DEVICES:
            return cast(Self, ClientNotificationObservedDevices.from_json(values))
        raise NotImplementedError


class ClientNotificationEpacketReceived(ClientNotification):
    TYPE = ClientNotification.Type.EPACKET_RECV

    def __init__(self, epacket: PacketReceived):
        self.epacket = epacket

    def to_json(self) -> dict:
        """Convert class to json dictionary"""
        return {"type": int(self.TYPE), "epacket": self.epacket.to_json()}

    @classmethod
    def from_json(cls, values: dict) -> Self:
        return cls(PacketReceived.from_json(values["epacket"]))


class ClientNotificationObservedDevices(ClientNotification):
    TYPE = ClientNotification.Type.KNOWN_DEVICES

    def __init__(self, devices: dict[int, dict]):
        self.devices = devices

    def to_json(self) -> dict:
        """Convert class to json dictionary"""
        return {"type": int(self.TYPE), "devices": json.dumps(self.devices)}

    @classmethod
    def from_json(cls, values: dict) -> Self:
        raw = json.loads(values["devices"])
        decoded = {int(k): v for k, v in raw.items()}
        return cls(decoded)


class ClientNotificationConnection(ClientNotification):
    TYPE = 0

    def __init__(self, infuse_id: int):
        self.infuse_id = infuse_id

    def to_json(self) -> dict:
        return {"type": int(self.TYPE), "infuse_id": self.infuse_id}

    @classmethod
    def from_json(cls, values: dict) -> Self:
        return cls(values["infuse_id"])


class ClientNotificationConnectionFailed(ClientNotificationConnection):
    """Connection to device failed"""

    TYPE = ClientNotificationConnection.Type.CONNECTION_FAILED


class ClientNotificationConnectionCreated(ClientNotificationConnection):
    """Connection to device has been created"""

    TYPE = ClientNotificationConnection.Type.CONNECTION_CREATED

    def __init__(self, infuse_id: int, max_payload: int):
        super().__init__(infuse_id)
        self.max_payload = max_payload

    def to_json(self) -> dict:
        return {"type": int(self.TYPE), "infuse_id": self.infuse_id, "max_payload": self.max_payload}

    @classmethod
    def from_json(cls, values: dict) -> Self:
        return cls(values["infuse_id"], values["max_payload"])


class ClientNotificationConnectionDropped(ClientNotificationConnection):
    """Connection to device has been lost"""

    TYPE = ClientNotificationConnection.Type.CONNECTION_DROPPED


class GatewayRequest:
    class Type(enum.IntEnum):
        EPACKET_SEND = 0
        CONNECTION_REQUEST = 1
        CONNECTION_RELEASE = 2
        KNOWN_DEVICES = 3

    def to_json(self) -> dict:
        """Convert class to json dictionary"""
        raise NotImplementedError

    @classmethod
    def from_json(cls, values: dict) -> Self:
        """Reconstruct class from json dictionary"""
        if values["type"] == cls.Type.EPACKET_SEND:
            return cast(Self, GatewayRequestEpacketSend.from_json(values))
        elif values["type"] == cls.Type.CONNECTION_REQUEST:
            return cast(Self, GatewayRequestConnectionRequest.from_json(values))
        elif values["type"] == cls.Type.CONNECTION_RELEASE:
            return cast(Self, GatewayRequestConnectionRelease.from_json(values))
        elif values["type"] == cls.Type.KNOWN_DEVICES:
            return cast(Self, GatewayRequestObservedDevices.from_json(values))
        raise NotImplementedError


class GatewayRequestEpacketSend(GatewayRequest):
    """Request packet to be forwarded to device"""

    TYPE = GatewayRequest.Type.EPACKET_SEND

    def __init__(self, epacket: PacketOutput):
        self.epacket = epacket

    def to_json(self) -> dict:
        return {"type": int(self.TYPE), "epacket": self.epacket.to_json()}

    @classmethod
    def from_json(cls, values: dict) -> Self:
        return cls(PacketOutput.from_json(values["epacket"]))


class GatewayRequestObservedDevices(GatewayRequest):
    """Request list of known devices"""

    TYPE = GatewayRequest.Type.KNOWN_DEVICES

    def __init__(self):
        pass

    def to_json(self) -> dict:
        return {"type": int(self.TYPE)}

    @classmethod
    def from_json(cls, values: dict) -> Self:
        return cls()


class GatewayRequestConnection(GatewayRequest):
    TYPE = 0

    def __init__(self, infuse_id: int):
        self.infuse_id = infuse_id

    def to_json(self) -> dict:
        return {"type": int(self.TYPE), "infuse_id": self.infuse_id}

    @classmethod
    def from_json(cls, values: dict) -> Self:
        return cls(values["infuse_id"])


class GatewayRequestConnectionRequest(GatewayRequestConnection):
    """Request connection context to device"""

    TYPE = GatewayRequestConnection.Type.CONNECTION_REQUEST

    class DataType(enum.IntFlag):
        COMMAND = 1
        DATA = 2
        LOGGING = 4

    def __init__(self, infuse_id: int, data_types: DataType, timeout_ms: int):
        super().__init__(infuse_id)
        self.data_types = data_types
        self.timeout_ms = timeout_ms

    def to_json(self) -> dict:
        return {
            "type": int(self.TYPE),
            "infuse_id": self.infuse_id,
            "data_types": self.data_types,
            "timeout": self.timeout_ms,
        }

    @classmethod
    def from_json(cls, values: dict) -> Self:
        return cls(values["infuse_id"], values["data_types"], values["timeout"])


class GatewayRequestConnectionRelease(GatewayRequestConnection):
    """Release connection context to device"""

    TYPE = GatewayRequestConnection.Type.CONNECTION_RELEASE


class LocalServer:
    def __init__(self, multicast_address):
        # Multicast output socket
        self._output_sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM, socket.IPPROTO_UDP)
        self._output_sock.setsockopt(socket.IPPROTO_IP, socket.IP_MULTICAST_TTL, 2)
        self._output_sock.setsockopt(socket.IPPROTO_IP, socket.IP_MULTICAST_IF, socket.inet_aton("127.0.0.1"))
        self._output_addr = multicast_address
        # Single input socket
        unicast_address = ("localhost", multicast_address[1] + 1)
        self._input_sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM, socket.IPPROTO_UDP)
        self._input_sock.bind(unicast_address)
        self._input_sock.settimeout(0.2)

    def broadcast(self, notification: ClientNotification):
        self._output_sock.sendto(json.dumps(notification.to_json()).encode("utf-8"), self._output_addr)

    def receive(self) -> GatewayRequest | None:
        try:
            data, _ = self._input_sock.recvfrom(8192)
        except TimeoutError:
            return None
        return GatewayRequest.from_json(json.loads(data.decode("utf-8")))

    def close(self):
        self._input_sock.close()
        self._output_sock.close()


class LocalClient:
    def __init__(self, multicast_address, rx_timeout=0.2):
        # Multicast input socket
        self._input_sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM, socket.IPPROTO_UDP)
        self._input_sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        if sys.platform == "win32":
            self._input_sock.bind(("", multicast_address[1]))
        else:
            self._input_sock.bind(multicast_address)
        mreq = struct.pack("4s4s", socket.inet_aton(multicast_address[0]), socket.inet_aton("127.0.0.1"))
        self._input_sock.setsockopt(socket.IPPROTO_IP, socket.IP_ADD_MEMBERSHIP, mreq)
        self._input_sock.settimeout(rx_timeout)
        # Unicast output socket
        self._output_addr = ("localhost", multicast_address[1] + 1)
        self._output_sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM, socket.IPPROTO_UDP)
        # Connection context
        self._connection_id = None

    def set_rx_timeout(self, timeout):
        self._input_sock.settimeout(timeout)

    def send(self, request: GatewayRequest):
        self._output_sock.sendto(json.dumps(request.to_json()).encode("utf-8"), self._output_addr)

    def receive(self) -> ClientNotification | None:
        try:
            data, _ = self._input_sock.recvfrom(8192)
        except TimeoutError:
            return None
        return ClientNotification.from_json(json.loads(data.decode("utf-8")))

    def connection_create(
        self, infuse_id: int, data_types: GatewayRequestConnectionRequest.DataType, timeout_ms: int
    ) -> int:
        self._connection_id = infuse_id

        # Send the request for the connection
        req = GatewayRequestConnectionRequest(infuse_id, data_types, timeout_ms)
        self.send(req)
        # Wait for response from the server
        while True:
            if rsp := self.receive():
                if isinstance(rsp, ClientNotificationConnectionCreated):
                    return rsp.max_payload
                elif isinstance(rsp, ClientNotificationConnectionFailed):
                    raise ConnectionRefusedError

    def connection_release(self):
        assert self._connection_id is not None

        req = GatewayRequestConnectionRelease(
            self._connection_id,
        )
        self.send(req)
        self._connection_id = None

    @contextmanager
    def connection(self, infuse_id: int, data_types: GatewayRequestConnectionRequest.DataType, timeout_ms: int = 10000):
        try:
            yield self.connection_create(infuse_id, data_types, timeout_ms)
        finally:
            self.connection_release()

    def close(self):
        # Cleanup any lingering connection context
        if self._connection_id:
            req = GatewayRequestConnectionRelease(
                self._connection_id,
            )
            self.send(req)
        # Close the socket
        self._input_sock.close()

    def observe_announce(self) -> Generator[tuple[HopReceived, readings.announce | readings.announce_v2], None, None]:
        decoder = TDF()
        while True:
            msg = self.receive()
            if msg is None:
                continue
            if not isinstance(msg, ClientNotificationEpacketReceived):
                continue
            if msg.epacket.ptype != InfuseType.TDF:
                continue
            source = msg.epacket.route[0]

            for tdf in decoder.decode(msg.epacket.payload):
                if isinstance(tdf.data[0], readings.announce):
                    yield (source, tdf.data[0])
                if isinstance(tdf.data[0], readings.announce_v2):
                    yield (source, tdf.data[0])



from collections.abc import Mapping
from typing import Any, TypeVar

from attrs import define as _attrs_define
from attrs import field as _attrs_field

from ..models.definitions_field_conversion_int import DefinitionsFieldConversionInt
from ..types import UNSET, Unset

T = TypeVar("T", bound="DefinitionsFieldConversion")


@_attrs_define
class DefinitionsFieldConversion:
    """Conversion formula for a field (m * <value> + c)

    Attributes:
        m (Union[Unset, float]):
        c (Union[Unset, float]):
        int_ (Union[Unset, DefinitionsFieldConversionInt]): Byte array value should be treated  as an integer
    """

    m: Unset | float = UNSET
    c: Unset | float = UNSET
    int_: Unset | DefinitionsFieldConversionInt = UNSET
    additional_properties: dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> dict[str, Any]:
        m = self.m

        c = self.c

        int_: Unset | str = UNSET
        if not isinstance(self.int_, Unset):
            int_ = self.int_.value

        field_dict: dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update({})
        if m is not UNSET:
            field_dict["m"] = m
        if c is not UNSET:
            field_dict["c"] = c
        if int_ is not UNSET:
            field_dict["int"] = int_

        return field_dict

    @classmethod
    def from_dict(cls: type[T], src_dict: Mapping[str, Any]) -> T:
        d = dict(src_dict)
        m = d.pop("m", UNSET)

        c = d.pop("c", UNSET)

        _int_ = d.pop("int", UNSET)
        int_: Unset | DefinitionsFieldConversionInt
        if isinstance(_int_, Unset):
            int_ = UNSET
        else:
            int_ = DefinitionsFieldConversionInt(_int_)

        definitions_field_conversion = cls(
            m=m,
            c=c,
            int_=int_,
        )

        definitions_field_conversion.additional_properties = d
        return definitions_field_conversion

    @property
    def additional_keys(self) -> list[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



#!/usr/bin/env python3

import ctypes
import sys
from http import HTTPStatus
from json import loads

import infuse_iot.definitions.rpc as defs
from infuse_iot.api_client import Client
from infuse_iot.api_client.api.coap import get_coap_file_stats
from infuse_iot.commands import InfuseRpcCommand
from infuse_iot.credentials import get_api_key
from infuse_iot.definitions.rpc import rpc_enum_file_action
from infuse_iot.util.ctypes import UINT32_MAX
from infuse_iot.zephyr.errno import errno


def coap_server_file_stats(server: str, resource: str) -> tuple[int, int]:
    if server == coap_download.INFUSE_COAP_SERVER_ADDR:
        # Validate file prefix
        if not resource.startswith("file/"):
            sys.exit("Infuse-IoT COAP files start with 'file/'")
        api_filename = resource.removeprefix("file/")
        # Get COAP file information
        client = Client(base_url="https://api.infuse-iot.com").with_headers({"x-api-key": f"Bearer {get_api_key()}"})
        with client as client:
            response = get_coap_file_stats.sync_detailed(client=client, filename=api_filename)
            decoded = loads(response.content.decode("utf-8"))
            if response.status_code != HTTPStatus.OK:
                sys.exit(f"<{response.status_code}>: {decoded['message']}")
            return (decoded["len"], decoded["crc"])
    else:
        # Unknown, let the COAP download automatically determine
        # This does mean that duplicate file are not detected
        print("Custom COAP server, duplicate file detection disabled")
        return (UINT32_MAX, UINT32_MAX)


class coap_download(InfuseRpcCommand, defs.coap_download):
    INFUSE_COAP_SERVER_ADDR = "coap.dev.infuse-iot.com"
    INFUSE_COAP_SERVER_PORT = 5684

    @classmethod
    def add_parser(cls, parser):
        parser.add_argument(
            "--server",
            type=str,
            default=cls.INFUSE_COAP_SERVER_ADDR,
            help="COAP server name",
        )
        parser.add_argument(
            "--port",
            type=int,
            default=cls.INFUSE_COAP_SERVER_PORT,
            help="COAP server port",
        )
        parser.add_argument(
            "--resource",
            "-r",
            type=str,
            required=True,
            help="Resource path",
        )
        group = parser.add_mutually_exclusive_group(required=True)
        group.add_argument(
            "--discard",
            dest="action",
            action="store_const",
            const=rpc_enum_file_action.DISCARD,
            help="Download file and discard without action",
        )
        group.add_argument(
            "--dfu",
            dest="action",
            action="store_const",
            const=rpc_enum_file_action.APP_IMG,
            help="Download complete image file and perform DFU",
        )
        group.add_argument(
            "--cpatch",
            dest="action",
            action="store_const",
            const=rpc_enum_file_action.APP_CPATCH,
            help="Download CPatch binary diff and perform DFU",
        )
        group.add_argument(
            "--nrf91-modem",
            dest="action",
            action="store_const",
            const=rpc_enum_file_action.NRF91_MODEM_DIFF,
            help="nRF91 LTE modem diff upgrade",
        )
        group.add_argument(
            "--for-copy",
            dest="action",
            action="store_const",
            const=rpc_enum_file_action.FILE_FOR_COPY,
            help="File to copy to other device",
        )

    def __init__(self, args):
        self.server = args.server.encode("utf-8")
        self.port = args.port
        self.resource = args.resource.encode("utf-8")
        self.action = args.action
        self.file_len, self.file_crc = coap_server_file_stats(args.server, args.resource)

    def request_struct(self):
        class request(ctypes.LittleEndianStructure):
            _fields_ = [
                ("server_address", 48 * ctypes.c_char),
                ("server_port", ctypes.c_uint16),
                ("block_timeout_ms", ctypes.c_uint16),
                ("action", ctypes.c_uint8),
                ("resource_len", ctypes.c_uint32),
                ("resource_crc", ctypes.c_uint32),
                ("resource", (len(self.resource) + 1) * ctypes.c_char),
            ]
            _pack_ = 1

        return request(
            self.server,
            self.port,
            2000,
            self.action,
            self.file_len,
            self.file_crc,
            self.resource,
        )

    def request_json(self):
        return {
            "server_address": self.server.decode("utf-8"),
            "server_port": str(self.port),
            "block_timeout_ms": "2000",
            "action": self.action.name,
            "resource_len": str(self.file_len),
            "resource_crc": str(self.file_crc),
            "resource": self.resource.decode("utf-8"),
        }

    def handle_response(self, return_code, response):
        if return_code != 0:
            print(f"Failed to download file ({errno.strerror(-return_code)})")
            return
        else:
            print("File downloaded")
            print(f"\tLength: {response.resource_len}")
            print(f"\t   CRC: 0x{response.resource_crc:08x}")
